/* wptKeyserver.h - Keyserver headerfile
 *	Copyright (C) 2000-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */

#ifndef WPT_KEYSERVER_H
#define WPT_KEYSERVER_H

#define HKP_PORT     11371 /* Horowitz Keyserver Protocol */
#define FINGER_PORT	79 /* Finger Protocol */
#define LDAP_PORT      389 /* Light Weight Dir Access Protocol. */

/* Default keyserver URL. */
#define DEF_HKP_KEYSERVER "http://subkeys.pgp.net"

/* Valid proxy protocols. */
enum proxy_proto_t {
    PROXY_PROTO_NONE = 0,
    PROXY_PROTO_HTTP,
    PROXY_PROTO_SOCKS4,
    PROXY_PROTO_SOCKS5
};

/* Valid proxy authentications. */
enum proxy_auth_t {
    PROXY_AUTH_NONE = 0,
    PROXY_AUTH_PLAIN = 2
};

/* Valid key server protocols. */
enum ksproto_t {
    KSPROTO_HTTP   = 0,
    KSPROTO_LDAP   = 1,
    KSPROTO_FINGER = 2
};
#define KSPROTO_HKP 0 /* alias for http */

struct keyserver_ctx {
    const char	  *name;
    unsigned short port;
    const char	  *pattern;
};

struct keyserver {
    char *name;	    /* name of the keyserver. */
    int   port;	    /* used tcp port. */
    int	  used;	    /* 1=entry will be used. */
    int	  proto;    /* protocol (0..2). */
};

/* Proxy configuration context. */
struct keyserver_proxy_s {
    int   proto;	/* protocol. */
    int	  port;		/* server port. */
    char *host;		/* server host. */
    int   auth_mode;	/* authentication mode. */
    char *user;		/* user. */
    char *base64_user;	/* base64 encoded user. */
    char *pass;		/* password. */
};
typedef struct keyserver_proxy_s *keyserver_proxy_t;

#define MAX_KEYSERVERS 32
extern keyserver server[MAX_KEYSERVERS]; /* runtime initialized */
extern char *default_keyserver;
extern WORD default_keyserver_port;
extern keyserver_proxy_s proxy;


struct keyserver_uid_s {
    struct keyserver_uid_s *next;
    long creation;
    long expires;
    char *uid;
};

struct keyserver_key_s {
    int	   algo;		    /* pubkey algorithm. */
    int	   bits;		    /* length of the key. */
    char   *keyid;		    /* keyid of the key. */
    long   creation;		    /* creation date. */
    long   expires;
    struct keyserver_uid_s *uids;   /* list ofuser-ids. */
    const struct keyserver_uid_s *main_uid;
};


/*-- wptSOCKS.cpp --*/
int socks_handshake (keyserver_proxy_t ctx, int conn_fd,
		     const char *hostname, WORD port);

/*-- Winsock2 support --*/
int wsock_init (void);
void wsock_end (void);
const char* wsock_strerror (void);

/*-- Keyserver config --*/
int kserver_load_conf (const char *conf);
int kserver_save_conf (const char *conf);
const char* kserver_check_keyid (const char *keyid);
void kserver_proxy_release (keyserver_proxy_t ctx);
const char* kserver_get_hostname (int idx, int type, unsigned short *port);
void keyserver_set_default (const char *hostname, unsigned short port);
void kserver_set_socket_timeout (int nsec);

/*-- General --*/
int check_IP_or_hostname (const char *name);
const char* kserver_strerror (void);
int kserver_check_inet_connection (void);
int kserver_connect (const char *hostname, WORD port, int *conn_fd);
int kserver_sendkey (const char *hostname, WORD port, 
		     const char *pubkey, int len);
int kserver_recvkey (const char *hostname, WORD port, 
		     const char *keyid, char **r_key, int *r_keylen);
int kserver_search_begin (const char *hostname, WORD port, 
			  const char *pattern, int *conn_fd, int *nkeys);
int kserver_search_next (int conn_fd, keyserver_key_s **r_key);
void kserver_search_end (int conn_fd);
void kserver_release_key (keyserver_key_s *key);

/*-- Finger/LDAP Protocol --*/
int  ldap_recvkey (const char *hostname, const char *keyid, 
		   char **r_key, int *r_keylen);
int finger_recvkey (const char *host, const char *user, 
		    char **r_key, int *r_keylen);

/*-- Dialog support --*/
int hkp_send_key (HWND dlg, const char *kserver, unsigned short port, 
		  const char *pattern);
int hkp_recv_key (HWND dlg, const char *kserver, unsigned short port, 
		  const char *pattern, int proto, int flags);

#endif /* WPT_KEYSERVER_H */
