/* wptFileMangerDlg.cpp - File Manager
 *	Copyright (C) 2001-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>
#include <sys/stat.h>
#include <stdio.h>

#include "resource.h"
#include "wptCommonCtl.h"
#include "wptGPG.h"
#include "wptW32API.h"
#include "wptKeylist.h"
#include "wptErrors.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptTypes.h"
#include "wptNLS.h"
#include "wptVersion.h"
#include "wptFileManager.h"
#include "wptRegistry.h"
#include "wptCrypto.h"

#define FM_SEPARATOR_ID 10001

#define send_cmd_id(hwnd, id) \
    PostMessage ((hwnd), WM_COMMAND, MAKEWPARAM( (id), 0), 0)

static subclass_s filelist_proc;

struct thread_ctx_s {
    HWND   dlg;
    HANDLE fd;
    char *drive;
};
typedef struct thread_ctx_s *thread_ctx_t;


int is_openpgp_ext (const char *name);

/* Return 1 if the selected file in the listview has a valid GPG extension. */
static int
is_openpgp_file (listview_ctrl_t lv)
{
    char name[MAX_PATH+32];

    listview_get_item_text (lv, listview_get_curr_pos (lv), 1, 
			    name, sizeof (name)-1);
    return is_openpgp_ext (name);
}


/* Dialog procedure for selecting recipients for encryption. */
BOOL CALLBACK
file_encrypt_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static fm_state_t c;
    static keylist_t list = NULL;
    gpg_keycache_t kc;
    gpgme_ctx_t ctx;
    int force_trust = 0;
    
    switch (msg) {
    case WM_INITDIALOG:
        c = (fm_state_t)lparam;
        if (!c)
	    BUG (0);
        kc = keycache_get_ctx (KEYCACHE_PUB);
        SetWindowText (dlg, _("File Encrypt"));
        SetDlgItemText (dlg, IDC_ENCRYPT_ARMOR, _("&Text Output"));
	SetDlgItemText (dlg, IDC_ENCRYPT_WIPE, _("&Wipe Original"));
	SetDlgItemText (dlg, IDC_ENCRYPT_ANON, _("Anonymize (throw keyid)"));
	SetDlgItemText (dlg, IDC_ENCRYPT_SIGN, _("&Select Key for signing"));
	SetDlgItemText (dlg, IDC_ENCRYPT_OPTINF, _("Options"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));

	/* XXX: support --hidden-recipient */
	EnableWindow (GetDlgItem (dlg, IDC_ENCRYPT_ANON), FALSE);

	if (c->req_signer) {
	    EnableWindow (GetDlgItem (dlg, IDC_ENCRYPT_SIGN), TRUE);
	    seclist_init (dlg, IDC_ENCRYPT_SECLIST, KEYLIST_FLAG_SHORT, &list);
	}
        lv = keylist_load (GetDlgItem (dlg, IDC_ENCRYPT_LIST), kc, NULL,
			   KEYLIST_ENCRYPT_MIN|KEYLIST_FLAG_FILE, 
			   KEY_SORT_USERID);
        SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_DESTROY:
	seclist_destroy (&list);
        if (lv) {
            keylist_delete (lv);
            lv = NULL;
        }
        return FALSE;

    case WM_NOTIFY:
	NMHDR *notify;
        notify = (NMHDR *)lparam;
	if (notify && notify->code == LVN_COLUMNCLICK 
	    && notify->idFrom == IDC_ENCRYPT_LIST) {
	    NMLISTVIEW *p = (LPNMLISTVIEW) lparam;
	    int sortby = 0;

	    switch (p->iSubItem) {
	    case 0: sortby = KEY_SORT_USERID; break;
	    case 1: sortby = KEY_SORT_KEYID; break;
	    case 2: sortby = KEY_SORT_LEN; break;
	    case 4: sortby = KEY_SORT_VALIDITY; break;
	    default:sortby = KEY_SORT_USERID; break;
	    }
	    keylist_sort (lv, sortby);
	}
	break;
        
    case WM_COMMAND:
	if (HIWORD (wparam) == BN_CLICKED
	    && LOWORD (wparam) == IDC_ENCRYPT_SIGN) {
	    int req_signer = IsDlgButtonChecked (dlg, IDC_ENCRYPT_SIGN);
	    EnableWindow (GetDlgItem (dlg, IDC_ENCRYPT_SECLIST),
			  req_signer? TRUE : FALSE);
	}
        switch (LOWORD (wparam)) {
        case IDOK:
            ctx = c->ctx;
	    safe_free (c->recp);
            c->recp = keylist_get_recipients (lv, &force_trust, &c->n_recp);
	    if (!c->n_recp) {
		msg_box (dlg, _("Please select at least one recipient."),
			 _("File Encrypt"), MB_ERR);
		safe_free (c->recp);
		c->recp = NULL;
                c->cancel = 1;
                return TRUE;
            }
	    if (IsDlgButtonChecked (dlg, IDC_ENCRYPT_SIGN)) {
		gpgme_key_t key;

		if (seclist_select_key (dlg, IDC_ENCRYPT_SECLIST, &key))
		    return TRUE;
		else
		    gpgme_signers_add (c->ctx, key);
	    }

            if (IsDlgButtonChecked (dlg, IDC_ENCRYPT_ARMOR))
		gpgme_set_armor (ctx, 1);
	    if (IsDlgButtonChecked (dlg, IDC_ENCRYPT_WIPE))
		c->wipe = 1;
	    /* XXX: implement this
	    if (IsDlgButtonChecked (dlg, IDC_ENCRYPT_ANON))
		;
	    */
            c->cancel = 0;
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            c->cancel = 1;
            EndDialog (dlg, FALSE);
            return TRUE;
        }
        return TRUE;
    }
    
    return FALSE;
}


/* Dialog box procedure for importing keys. */
BOOL CALLBACK
file_import_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static fm_state_t c;
    static int is_revcert = 0, has_seckeys = 0;
    
    switch (msg) {
    case WM_INITDIALOG:
        c = (fm_state_s *)lparam;
        if (!c)
            BUG (0);
        ShowWindow (GetDlgItem (dlg, IDC_IMPORT_KEYMNGT), WS_DISABLED);
        implist_build (&lv, GetDlgItem (dlg, IDC_IMPORT_KEYLIST));
        implist_load (lv, c->import.is_clip? NULL: c->opaque, 
		      &is_revcert, &has_seckeys);
	if (!listview_count_items (lv, 0)) {
	    msg_box (dlg, _("No valid OpenPGP data found."), _("Key Import"), MB_ERR);
	    c->cancel = 1;
	    EndDialog (dlg, FALSE);
	    return TRUE;
	}
	else {
	    const char *s = _("%s contain(s) %d key(s).");
	    char *p = new char[strlen (s) + 32];
	    if (!p)
		BUG (NULL);
	    sprintf (p, s, c->import.is_clip? _("Clipboard") : _("File"),
		     listview_count_items (lv, 0));
	    SetDlgItemText (dlg, IDC_IMPORT_INFO, p);
	    free_if_alloc (p);
	}
	SetWindowText (dlg, _("Key Import"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            listview_release (lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDC_IMPORT_DOIT:
	    if (has_seckeys > 0) {
		msg_box (dlg, _("Some of the imported keys are secret keys.\n\n"
				"The ownertrust values of these keys must be\n"
				"set manually via the Key Properties dialog."),
			 _("Import"), MB_INFO);
            }
            c->cancel = 0;
	    c->import.revcert = is_revcert? 1 : 0;
	    c->import.has_seckey = has_seckeys? 1 : 0;
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            c->cancel = 1;
            EndDialog (dlg, FALSE);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}


/* Dialog box procedure for file signing. */
BOOL CALLBACK
file_sign_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static fm_state_t c;
    static listview_ctrl_t lv;
    gpg_keycache_t kc, sec_kc;    
    gpgme_key_t key;
    int lv_idx;
    char signer[32];
    
    switch (msg) {
    case WM_INITDIALOG:
        c = (fm_state_s *)lparam;
	if (!c)
	    BUG (0);
        SetWindowText (dlg, _("File Sign"));
        SetDlgItemText (dlg, IDC_FILE_SIG_ARMOR, _("&Text Output") );
        SetDlgItemText (dlg, IDC_FILE_SIG_NORMAL, _("&Normal Signature"));
        SetDlgItemText (dlg, IDC_FILE_SIG_DETACH, _("&Detached Signature"));
        SetDlgItemText (dlg, IDC_FILE_SIG_CLEAR, _("&Cleartext Signature"));
	SetDlgItemText (dlg, IDC_FILE_SIG_SIGINF, _("Signature mode"));
	SetDlgItemText (dlg, IDC_FILE_SIG_OPTINF, _("Options"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	CheckDlgButton (dlg, IDC_FILE_SIG_DETACH, BST_CHECKED);
        kc = keycache_get_ctx (KEYCACHE_PUB);
        if (!kc)
            BUG (dlg);
        sec_kc = keycache_get_ctx (KEYCACHE_PRV);
	if (!sec_kc)
	    BUG (dlg);
	lv = keylist_load (GetDlgItem (dlg, IDC_FILE_SIG_LIST), kc, sec_kc,
			    KEYLIST_SIGN, KEY_SORT_USERID);
        SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            listview_release (lv);
            lv = NULL;
        }
        return FALSE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDOK:
            if (IsDlgButtonChecked (dlg, IDC_FILE_SIG_ARMOR))
		gpgme_set_armor (c->ctx, 1);
            if (IsDlgButtonChecked (dlg, IDC_FILE_SIG_NORMAL))
                c->sigmode = GPGME_SIG_MODE_NORMAL;
            else if (IsDlgButtonChecked (dlg, IDC_FILE_SIG_DETACH))
                c->sigmode = GPGME_SIG_MODE_DETACH;
            else if (IsDlgButtonChecked( dlg, IDC_FILE_SIG_CLEAR))
                c->sigmode = GPGME_SIG_MODE_CLEAR;
            else {
                /* default is to produce detached signatures */
                c->sigmode = GPGME_SIG_MODE_DETACH;
            }
            lv_idx = listview_get_selected_item (lv);
            if (lv_idx == -1) {
		if (listview_count_items (lv, 0) == 1)
		    lv_idx = 0;
		else {
		    msg_box (dlg, _("Please select a key."), _("File Sign"), MB_ERR);
		    return FALSE;
		}
	    }
            listview_get_item_text (lv, lv_idx, 1, signer, sizeof (signer)-1);
	    if (get_seckey (signer, &key))
		BUG (NULL);
	    gpgme_signers_clear (c->ctx);
	    gpgme_signers_add (c->ctx, key);
	    free_if_alloc (c->opaque);
	    c->opaque = m_strdup (signer); /* store for later use */
            c->cancel = 0;
            EndDialog (dlg, TRUE);
            return TRUE;
            
        case IDCANCEL:
            c->cancel = 1;
            EndDialog (dlg, FALSE);
            return TRUE;
        }
        break;
    }
    
    return FALSE;
}


/* Setup status bar for the main window @dlg. */
static HWND
setup_status_bar (HWND dlg)
{
    HWND statbar;
    RECT r;
    int partpos[3];
    int i;

    GetClientRect (dlg, &r);
    for (i=1; i <= 3; i++)
	partpos[i-1] = r.right/3*i;

    statbar = CreateStatusWindow (WS_CHILD | WS_VISIBLE | CCS_BOTTOM, 
				   "", dlg, FM_SEPARATOR_ID);
    ShowWindow (statbar, SW_SHOW);
    SendMessage (statbar, SB_SETPARTS, (WPARAM)3, (LPARAM)partpos);
    return statbar;
}


/* Refresh the status bar text with the selected item from @lv. */
static void
update_status_bar (HWND statbar, listview_ctrl_t lv)
{
    struct stat st;
    char buf[128];
    int i, n;
    DWORD len = 0;

    if (!lv) { /* flush */
	SendMessage (statbar, SB_SETTEXT, 1, (LPARAM)"0 Byte");
	SendMessage (statbar, SB_SETTEXT, 0, (LPARAM)_("0 Objects marked"));
	return;
    }
    n = listview_count_items (lv, 1);
    _snprintf (buf, sizeof (buf)-1, _("%d Object(s) marked"), n);
    SendMessage (statbar, SB_SETTEXT, 0, (LPARAM)buf);
    if (!n)
	return;
    for (i=0; i < listview_count_items (lv, 0); i++) {
	if (!listview_get_item_state (lv, i))
	    continue;
	listview_get_item_text (lv, i, 1, buf, sizeof (buf)-1);
	if (stat (buf, &st) == 0)
	    len += st.st_size;
    }
    if (!len)
	_snprintf (buf, sizeof (buf)-1, "0 Byte");
    else if (len < 1024)
	_snprintf (buf, sizeof (buf)-1, "%d Byte", len);
    else if (st.st_size < 1048576)
	_snprintf (buf, sizeof (buf)-1, "%d KB", len/1024);
    else
	_snprintf (buf, sizeof (buf)-1, "%d MB", len/1024/1024);
    SendMessage (statbar, SB_SETTEXT, 1, (LPARAM)buf);
}


/* Subclass procedure to detect common hotkeys. */
static BOOL CALLBACK
filelist_subclass_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    int virt_key;

    switch (msg) {
    case WM_KEYUP:
	virt_key = (int)wparam;
	if (virt_key == VK_DELETE)
	    send_cmd_id (filelist_proc.dlg, ID_FILEMISC_WIPE);
	break;
    }
    return CallWindowProc (filelist_proc.old, dlg, msg, wparam, lparam);
}


/* Enable or disabled menu entries from the dialog window @hwnd.
   @lv is used to find out the state of the selected file entry. */
static void
update_ui_items (HWND hwnd, HMENU hm, bool is_popup, listview_ctrl_t lv)
{
    int ids[2][7] = {
	{ID_FILEMISC_SEND, ID_FILEMISC_DECRYPT, ID_FILEMISC_VERIFY,
	 ID_FILEMISC_ENCRYPT, ID_FILEMISC_SIGN, ID_FILEMISC_SIGNENC,
	 ID_FILEMISC_ENCRYPT_ZIP
	},
	{ID_FILECTX_SEND,  ID_FILECTX_DECRYPT,  ID_FILECTX_VERIFY,
	 ID_FILECTX_ENCRYPT, ID_FILECTX_SIGN,   ID_FILECTX_SIGNENC,
	 ID_FILECTX_ENCRYPT_ZIP
	}
    };
    enum id_type_t { SEND, DECRYPT, VERIFY, ENCR, SIGN, SIGENC, ENCRZIP };
    int pos, idx = is_popup? 1: 0;

    pos = listview_get_curr_pos (lv);
    if (fm_check_file_type (lv, pos, FM_DECRYPT))
	set_menu_state (hm, ids[idx][SEND], MF_ENABLED);	

    if (fm_check_file_type (lv, pos, FM_ENCRYPT)) {	
	set_menu_state (hm, ids[idx][DECRYPT], MF_GRAYED|MF_DISABLED);	
	set_menu_state (hm, ids[idx][VERIFY], MF_GRAYED|MF_DISABLED);
	set_menu_state (hm, ids[idx][ENCR], MF_ENABLED);
	set_menu_state (hm, ids[idx][SIGN], MF_ENABLED);
	set_menu_state (hm, ids[idx][SIGENC], MF_ENABLED);
    }
    if (fm_check_file_type (lv, pos, FM_DECRYPT)) {
	set_menu_state (hm, ids[idx][ENCR], MF_GRAYED|MF_DISABLED);
	set_menu_state (hm, ids[idx][ENCRZIP], MF_GRAYED|MF_DISABLED);
	set_menu_state (hm, ids[idx][SIGN], MF_GRAYED|MF_DISABLED);
	set_menu_state (hm, ids[idx][SIGENC], MF_GRAYED|MF_DISABLED);
	set_menu_state (hm, ids[idx][DECRYPT], MF_ENABLED);
    }	    
    if (fm_check_file_type (lv, pos, FM_VERIFY)) {
	set_menu_state (hm, ids[idx][VERIFY], MF_ENABLED);
	set_menu_state (hm, ids[idx][SIGN], MF_GRAYED|MF_DISABLED);	
	set_menu_state (hm, ids[idx][SIGENC], MF_GRAYED|MF_DISABLED);
    }

    /* disable all operation which need a secret key. */
    if (!secret_key_available ()) {
	set_menu_state (hm, ids[idx][DECRYPT], MF_GRAYED|MF_DISABLED);
	set_menu_state (hm, ids[idx][SIGN], MF_GRAYED|MF_DISABLED);	
	set_menu_state (hm, ids[idx][SIGENC], MF_GRAYED|MF_DISABLED);
    }
}


/* Dialog box procedure for the File Manager. */
BOOL CALLBACK
file_manager_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static fm_info_t fm = NULL;
    static int always_ontop = 0;
    HWND fl;
    
    switch (msg) {
    case WM_INITDIALOG:
	always_ontop = get_reg_winpt_flag ("AlwaysOnTop");
	if (always_ontop == -1)
	    always_ontop = 0;
	SetWindowText (dlg, _("File Manager (use drag & drop to add files)"));
	fm_build (&fm, GetDlgItem (dlg, IDC_FILE_LIST));
	fm->menu = LoadMenu (glob_hinst, (LPCSTR)IDR_WINPT_FILEMISC);
	set_menu_text_bypos (fm->menu, 0, _("File"));
	set_menu_text_bypos (fm->menu, 1, _("Edit"));
	set_menu_text_bypos (fm->menu, 2, _("View"));
        set_menu_text (fm->menu, ID_FILEMISC_OPEN, _("&Open...") );
        set_menu_text (fm->menu, ID_FILEMISC_ENCRYPT, _("&Encrypt") );
	set_menu_text (fm->menu, ID_FILEMISC_ENCRYPT_ZIP, _("Encrypt into ZIP"));
        set_menu_text (fm->menu, ID_FILEMISC_DECRYPT, _("&Decrypt") );
        set_menu_text (fm->menu, ID_FILEMISC_SIGN, _("&Sign") );
        set_menu_text (fm->menu, ID_FILEMISC_SIGNENC, _("Sign && Encrypt") );
        set_menu_text (fm->menu, ID_FILEMISC_VERIFY, _("&Verify") );        
        set_menu_text (fm->menu, ID_FILEMISC_SYMENC, _("S&ymmetric") );
        set_menu_text (fm->menu, ID_FILEMISC_IMPORT, _("&Import") );
        set_menu_text (fm->menu, ID_FILEMISC_EXPORT, _("E&xport") );
        set_menu_text (fm->menu, ID_FILEMISC_QUIT, _("Exit"));
        set_menu_text (fm->menu, ID_FILEMISC_CLEAR, _("&Reset") );
	set_menu_text (fm->menu, ID_FILEMISC_ONTOP, _("Always on Top"));
	set_menu_text (fm->menu, ID_FILEMISC_PASTE, _("&Paste"));
	set_menu_text (fm->menu, ID_FILEMISC_SELALL, _("&Select All"));
	set_menu_text (fm->menu, ID_FILEMISC_PREFS, _("&Preferences"));
	set_menu_text (fm->menu, ID_FILEMISC_SEND, _("Send as Mail"));
	set_menu_text (fm->menu, ID_FILEMISC_LIST, _("&List Packets"));
	set_menu_text (fm->menu, ID_FILEMISC_WIPE_FREES, _("Wipe Free Space"));
	set_menu_text (fm->menu, ID_FILEMISC_WIPE, _("&Wipe"));	

	fl = GetDlgItem (dlg, IDC_FILE_LIST);
	filelist_proc.dlg = dlg;
	filelist_proc.current = (WNDPROC)filelist_subclass_proc;
	filelist_proc.old = (WNDPROC)GetWindowLong (fl, GWL_WNDPROC);
	if (filelist_proc.old) {
	    if (!SetWindowLong (fl, GWL_WNDPROC, (LONG)filelist_proc.current)) {
		msg_box (dlg, "Could not set filelist window procedure.", 
			 _("File Manager"), MB_ERR);
		BUG (NULL);
	    }
	}
	SendMessage (dlg, WM_SETICON, (WPARAM)ICON_SMALL, 
		     (LPARAM)LoadIcon (glob_hinst, (LPCTSTR)IDI_WINPT));
        SetMenu (dlg, fm->menu);
	if (!always_ontop)
	    set_menu_state (fm->menu, ID_FILEMISC_ONTOP, MFS_UNCHECKED);
	else
	    center_window2 (dlg, NULL, HWND_TOPMOST);
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
	fm->statbar = setup_status_bar (dlg);
	filelist_proc.opaque = (void*)fm->statbar;
        return TRUE;

    case WM_DESTROY:
        if (fm) {
	    fm_delete (fm);
            fm = NULL;
        }
	set_reg_winpt_flag ("AlwaysOnTop", always_ontop);
        return FALSE;
        
    case WM_DROPFILES:
        fm_add_dropped_files (fm, (HDROP)wparam);
        return TRUE;
        
    case WM_NOTIFY:
        NMHDR *notify;
	int pos;
	HMENU hm;	

        notify = (NMHDR *)lparam;
	if (notify && notify->code == LVN_COLUMNCLICK 
	    && notify->idFrom == IDC_FILE_LIST) {
	    NMLISTVIEW *p = (LPNMLISTVIEW) lparam;
	    fm_sort (fm->lv, p->iSubItem);
	}

	if (notify && notify->code == NM_CLICK &&
	    notify->idFrom == IDC_FILE_LIST)
	    update_status_bar (fm->statbar, fm->lv);
        if (notify && notify->code == NM_RCLICK
            && notify->idFrom == IDC_FILE_LIST) {
            POINT p;
            HMENU popup;
            
            GetCursorPos (&p);
            hm = LoadMenu (glob_hinst, MAKEINTRESOURCE (IDR_WINPT_FILEMISC_CTX));
            popup = GetSubMenu (hm, 0);
	    set_menu_text (popup, ID_FILECTX_ENCRYPT_ZIP, _("Encrypt into ZIP"));
            set_menu_text (popup, ID_FILECTX_ENCRYPT, _("Encrypt"));
            set_menu_text (popup, ID_FILECTX_DECRYPT, _("Decrypt"));
            set_menu_text (popup, ID_FILECTX_SIGN, _("Sign"));
            set_menu_text (popup, ID_FILECTX_SIGNENC, _("Sign && Encrypt"));
            set_menu_text (popup, ID_FILECTX_VERIFY, _("Verify"));
            set_menu_text (popup, ID_FILECTX_WIPE, _("Wipe"));
	    set_menu_text (popup, ID_FILECTX_SEND, _("Send as Mail"));
	    set_menu_text (popup, ID_FILECTX_LIST, _("List Packets"));
	    if (!is_openpgp_file (fm->lv))
		set_menu_state (popup, ID_FILECTX_LIST, MF_GRAYED|MF_DISABLED);
	    update_ui_items (dlg, popup, true, fm->lv);
            TrackPopupMenu (popup, TPM_RIGHTALIGN, p.x, p.y, 0, dlg, NULL);
            DestroyMenu (popup);
            DestroyMenu (hm);
        }
	if (notify && notify->code == NM_CLICK
            && notify->idFrom == IDC_FILE_LIST) {
	    hm = GetMenu (dlg);
	    pos = listview_get_curr_pos (fm->lv);
	    if (!fm_check_file_type (fm->lv, pos, FM_DECRYPT))
		set_menu_state (hm, ID_FILEMISC_SEND, MF_GRAYED|MF_DISABLED);
	    else
		set_menu_state (hm, ID_FILEMISC_SEND, MF_ENABLED);
	}
	if (notify && notify->code == LVN_ITEMCHANGED &&
            ((LPNMLISTVIEW)lparam)->uNewState) { /* item selected? */
	    update_ui_items (dlg, GetMenu (dlg), false, fm->lv);
	    update_status_bar (fm->statbar, fm->lv);
	}
        return TRUE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;
        
    case WM_INITMENUPOPUP:	
	if ((UINT)LOWORD (lparam) == 0 &&
	    GetMenuItemCount ((HMENU)wparam) == 19) /* XXX: do not hard code values */
	    set_menu_text_bypos ((HMENU)wparam, 2, _("&Calc Digest"));
	return FALSE;

    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case ID_FILEMISC_QUIT:
            EndDialog (dlg, TRUE);
            return TRUE;

	case ID_FILEMISC_WIPE_FREES:
	    if (!admin_user) {
		msg_box (dlg, _("This command requires admin privileges.\n"),
			 _("File Manager"), MB_ERR);
		break;
	    }
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_SPACE_SECDEL, 
			    dlg, space_wipefrees_dlg_proc, 0);
	    break;
            
	case ID_FILEMISC_SELALL:
	    listview_select_all (fm->lv);
	    break;

	case ID_FILEMISC_PASTE:
	    fm_verify_pasted_detsig (fm->lv, dlg);
	    break;

	case ID_FILEMISC_ONTOP:
	    always_ontop ^= 1;	    		
	    set_menu_state (fm->menu, ID_FILEMISC_ONTOP, 
			    always_ontop == 0? MFS_UNCHECKED : MFS_CHECKED);
	    center_window2 (dlg, NULL, 
			    always_ontop == 0? HWND_NOTOPMOST : HWND_TOPMOST);
	    break;

	case ID_FILEMISC_SEND:
	    fm_send_file (fm->lv);
	    break;

	case ID_FILEMISC_PREFS:
	    DialogBoxParam (glob_hinst, (LPCTSTR)IDD_FILE_PREFS, dlg,
			    file_preferences_dlg_proc, 0);
	    break;

        case ID_FILEMISC_LIST:
        case ID_FILEMISC_ENCRYPT:
        case ID_FILEMISC_SYMENC:
        case ID_FILEMISC_DECRYPT:
        case ID_FILEMISC_SIGN:
        case ID_FILEMISC_SIGNENC:
        case ID_FILEMISC_VERIFY:
        case ID_FILEMISC_WIPE:
	case ID_FILEMISC_ENCRYPT_ZIP:
            fm_parse_files (fm->lv, dlg, LOWORD (wparam));
	    SetForegroundWindow (dlg);
            return TRUE;
	    
	case ID_FILEMISC_MD_MD5:
	case ID_FILEMISC_MD_SHA1:
	case ID_FILEMISC_MD_RMD160:
	case ID_FILEMISC_MD_SHA256:
	    int algo;

	    switch (LOWORD (wparam)) {
	    case ID_FILEMISC_MD_MD5:	algo = GPGME_MD_MD5; break;
	    case ID_FILEMISC_MD_SHA1:	algo = GPGME_MD_SHA1; break;
	    case ID_FILEMISC_MD_RMD160: algo = GPGME_MD_RMD160; break; 
	    case ID_FILEMISC_MD_SHA256:	algo = GPGME_MD_SHA256; break;
	    default:			algo = GPGME_MD_SHA1; break;
	    }
	    fm_print_md (fm->lv, dlg, algo);
	    SetForegroundWindow (dlg);
	    return TRUE;
            
        case ID_FILEMISC_CLEAR:
            fm_reset (fm);
	    update_status_bar (fm->statbar, NULL);
            return TRUE;
            
        case ID_FILEMISC_OPEN:
            fm_add_opened_files (fm, dlg);
            return TRUE;
            
            /* context menu entries */
        case ID_FILECTX_ENCRYPT: send_cmd_id (dlg, ID_FILEMISC_ENCRYPT); break;
        case ID_FILECTX_DECRYPT: send_cmd_id (dlg, ID_FILEMISC_DECRYPT);break;
        case ID_FILECTX_SIGN:    send_cmd_id (dlg, ID_FILEMISC_SIGN);   break;
        case ID_FILECTX_SIGNENC: send_cmd_id (dlg, ID_FILEMISC_SIGNENC);break;
        case ID_FILECTX_VERIFY:  send_cmd_id (dlg, ID_FILEMISC_VERIFY); break;
        case ID_FILECTX_WIPE:    send_cmd_id (dlg, ID_FILEMISC_WIPE);   break;
        case ID_FILECTX_LIST:    send_cmd_id (dlg, ID_FILEMISC_LIST);   break;
	case ID_FILECTX_SEND:    send_cmd_id (dlg, ID_FILEMISC_SEND);   break;
	case ID_FILECTX_ENCRYPT_ZIP:send_cmd_id (dlg, ID_FILEMISC_ENCRYPT_ZIP); break;
        }
        break;
    }
    
    return FALSE;
}


/* Dialog box procedure for the FM preferences. */
BOOL CALLBACK
file_preferences_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    switch (msg) {
    case WM_INITDIALOG:
	CheckDlgButton (dlg, IDC_FPREFS_PROGRESS, 
			reg_prefs.fm.progress? BST_CHECKED : BST_UNCHECKED);
	SetForegroundWindow (dlg);
	center_window (dlg, NULL);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    reg_prefs.fm.progress = IsDlgButtonChecked (dlg, IDC_FPREFS_PROGRESS)? 1 : 0;
	    set_reg_winpt_prefs (&reg_prefs);
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    EndDialog (dlg, FALSE);
	    break;
	}
	break;
    }

    return FALSE;
}


/* Add the selected items from @lv into the combo box. */
static void 
add_files_from_lv (HWND dlg, listview_ctrl_t lv)
{
    char fname[384+1];
    int i;

    for (i = 0;  i < listview_count_items (lv, 0); i++) {
        if (!listview_get_item_state (lv, i))
            continue;	    
	listview_get_item_text (lv, i, 1, fname, sizeof (fname)-1);
	SendDlgItemMessage (dlg, IDC_SECDEL_FILES, LB_ADDSTRING,
			    0, (WPARAM)(char *)fname);
    }
}


/* Dialog box procedure to confirm that files shall be deleted. */
BOOL CALLBACK
file_secdel_confirm_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static secdel_confirm_s * ctx;

    switch (msg) {
    case WM_INITDIALOG:
	ctx = (struct secdel_confirm_s *)lparam;
	if (!ctx)
	    BUG (NULL);
	add_files_from_lv (dlg, ctx->lv_files);
	SetWindowText (dlg, _("Are you sure you want to secure delete these files?"));
	SetDlgItemText (dlg, IDYES, _("&Yes"));
	SetDlgItemText (dlg, IDNO, _("&No"));
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDYES:
	    ctx->yes = 1;
	    EndDialog (dlg, TRUE);
	    break;

	case IDNO:
	    ctx->yes = 0;
	    EndDialog (dlg, FALSE);
	    break;
	}
	break;
    }

    return FALSE;
}


/* Wipe callback. @ctx is a window handle @off is the current offset
   and @total is the total size. */
static void
wipe_cb (void *ctx, ui64 off, ui64 total)
{    
    SendDlgItemMessage ((HWND)ctx, IDC_SPACE_PROGRESS, PBM_SETPOS, 
			(WPARAM) (off*100/total), 0);
    SetDlgItemInt ((HWND)ctx, IDC_SPACE_OFF, (UINT)(off/1024), FALSE);
    SetDlgItemInt ((HWND)ctx, IDC_SPACE_N, (UINT)(total/1024), FALSE);
}


/* Call the wipe free space method asynchron. */
static DWORD WINAPI
wipe_thread (void *ctx)
{     
    thread_ctx_t a = (thread_ctx_t)ctx;
    int rc;

    rc = wipe_freespace (a->drive, &a->fd, wipe_cb, a->dlg);
    SetDlgItemText (a->dlg, IDC_SPACE_INFO, 
		    rc? _("Operation Status: Error") : 
		        _("Operation Status: Done."));
    if (rc)
	msg_box (a->dlg, winpt_strerror (rc), _("Wipe Free Space"), MB_ERR);
    else
	MessageBeep (MB_OK);
    ExitThread (0);
    return 0;
}


static void
fill_drive_combobox (HWND dlg)
{
    char buf[384], p[16];
    int n, i, j = 0;

    n = GetLogicalDriveStrings (DIM (buf)-1, buf);
    while (j < n) {
	for (i = 0; buf[j] != '\0'; i++, j++)
	    p[i] = buf[j];
	p[i] = '\0'; 
	j++;
	if (GetDriveType (p) != DRIVE_CDROM)
	    SendDlgItemMessage (dlg, IDC_SPACE_DRIVES, CB_ADDSTRING, 
				0, (LPARAM)(const char *)p);
    }
    SendDlgItemMessage (dlg, IDC_SPACE_DRIVES, CB_SETCURSEL, 0, 0);
}


static void
update_disc_stats (HWND dlg)
{    
    ULARGE_INTEGER caller = {0}, total = {0}, free = {0};
    char buf[128], drive[32], a_total[64] = {0}, a_free[64] = {0};
    int idx;
    
    idx = SendDlgItemMessage (dlg, IDC_SPACE_DRIVES, CB_GETCURSEL, 0, 0);
    GetDlgItemText (dlg, IDC_SPACE_DRIVES, drive, DIM (drive)-1);
    if (GetDiskFreeSpaceEx (drive, &caller, &total, &free)) {
	_ui64toa (total.QuadPart/1024, a_total, 10);
	_ui64toa (free.QuadPart/1024, a_free, 10);
    }
    _snprintf (buf, DIM (buf)-1, _("Total Capacity: %12sk\n"
			           "Free Space    : %12sk"), a_total, a_free);
    SetDlgItemText(dlg, IDC_SPACE_DISC_INFO, buf);
    SetDlgItemInt (dlg, IDC_SPACE_OFF, 0, FALSE);
    SetDlgItemInt (dlg, IDC_SPACE_N, 0, FALSE);
}


BOOL CALLBACK
space_wipefrees_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static HANDLE thread_hd = NULL;
    static thread_ctx_s ctx;
    char drive[32], * p;
    int idx;
    DWORD tid;

    switch (msg) {
    case WM_INITDIALOG:	
	fill_drive_combobox (dlg);
	update_disc_stats (dlg);
	center_window (dlg, NULL);
	SetForegroundWindow (dlg);
	break; 
	
    case WM_DESTROY:
	thread_hd = NULL;
	free_if_alloc (ctx.drive);
	memset (&ctx, 0, sizeof ctx);
	break;

    case WM_COMMAND:
	if (HIWORD (wparam) == CBN_SELCHANGE)
	    update_disc_stats (dlg);
	switch (LOWORD (wparam)) {
	case IDOK:
	    idx = SendDlgItemMessage (dlg, IDC_SPACE_DRIVES, CB_GETCURSEL, 0, 0);
	    if (idx != CB_ERR && !thread_hd) {
		GetDlgItemText (dlg, IDC_SPACE_DRIVES, drive, DIM (drive)-1);
		ctx.dlg = dlg;
		ctx.drive = m_strdup (drive);
		ctx.fd = NULL;
		thread_hd = CreateThread (NULL, 0, wipe_thread, &ctx, 0, &tid);
	    }
	    if (WaitForSingleObject (thread_hd, 0) == WAIT_OBJECT_0) {
		CloseHandle (thread_hd);
		thread_hd = NULL;
		EndDialog (dlg, TRUE);
	    }
	    break;

	case IDCANCEL:
	    if (thread_hd) {
		WaitForSingleObject (thread_hd, 2000);
		TerminateThread (thread_hd, 1);
		CloseHandle (thread_hd);
		CloseHandle (ctx.fd);
		p = make_filename (ctx.drive, "temp_winpt.tmp", NULL);
		if (p && !file_exist_check (p))
		    remove (p);
		free_if_alloc (p);
	    }
	    EndDialog( dlg, FALSE );
	    break;
	}
	break;
    }

    return FALSE;
}
