/* wptGPGME.cpp - WinPT GPGME interface
 *	Copyright (C) 2001-2006 Timo Schulz
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License 
 * as published by the Free Software Foundation; either version 2 
 * of the License, or (at your option) any later version.
 *  
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA 
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <sys/types.h>
#include <windows.h>

#include "resource.h"
#include "wptNLS.h"
#include "wptGPG.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptRegistry.h"
#include "wptDlgs.h"

#include "openpgp.h"

BOOL CALLBACK  keycache_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam);
void progress_cleanup (progress_filter_s * pfx);

/* Global GPG key cache contexts. */
static gpg_keycache_t pub = NULL;
static gpg_keycache_t sec = NULL;
static char *gpg_secring = NULL;


int
keycache_not_available (void)
{
    return pub == NULL;
}

/* Release both key cache objects. If @cleanup is 1,
   also release other global structs. */
void
keycache_release (int cleanup)
{
    char tmpbuf[64];
    int n = gpg_keycache_get_size (pub);

    /* XXX: update the value when the cache has changed. */
    sprintf (tmpbuf, "%d", n);
    set_reg_key (HKEY_CURRENT_USER, "Software\\WinPT", "nKeys", tmpbuf);

    if (pub) {
        gpg_keycache_release (pub); 
        pub = NULL;
    }
    if (sec) {
        gpg_keycache_release (sec);
        sec = NULL;
    }
    if (cleanup)
	safe_free (gpg_secring);
}


/* Update the key with the keyid @keyid in the key cache.
   If @is_sec is 1, the secret key cache is used. */
gpgme_error_t
keycache_update (int is_sec, const char *keyid)
{
    gpg_keycache_t ctx = pub;
    gpgme_error_t err;

    if (is_sec)
	ctx = sec;
    err = gpg_keycache_update_key (ctx, is_sec, pub, keyid);
    if (is_sec)
	gpg_keycache_prepare_single (ctx, keyid, NULL, gpg_secring);
    return err;
}


/* Initialize both cache contexts. Use @pubring for the public
   keyring and @secring for the secret keyring. */
gpgme_error_t
keycache_init (const char *pubring, const char *secring)
{
    struct progress_filter_s pfx;
    gpgme_error_t err;
    int val = 0;
    char *p;

    if (secring != NULL) {
	free_if_alloc (gpg_secring);
	gpg_secring = get_gnupg_keyring (0, NO_STRICT);
    }
    
    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "nKeys");
    if (p && *p != ' ') {
	val = atoi (p);
	free_if_alloc (p);	
    }

    memset (&pfx, 0, sizeof (pfx));
    keycache_release (0); /* Release old contexts first. */

    err = gpg_keycache_new (&pub);
    if (err)
	return err;
    if (val != 0)
	gpg_keycache_set_cb (pub, progress_callback, &pfx, val);
    err = gpg_keycache_new (&sec);
    if (!err)
	err = gpg_keycache_init (pub, NULL, 0);
    if (!err)
	err = gpg_keycache_init (sec, NULL, 1);
    if (!err && pubring && *pubring)
	err = gpg_keycache_prepare (pub, pubring, NULL);
    if (!err && secring && * secring)
	err = gpg_keycache_prepare (sec, NULL, secring);
    if (!err)
	gpg_keycache_sync (pub, sec);
    if (val != 0)
	progress_cleanup (&pfx);
    return err;
}


/* Return the public cache context if @is_pub is set
   the secre cache context otherwise. */
gpg_keycache_t
keycache_get_ctx (int is_pub)
{
    gpg_keycache_t ctx;

    ctx = is_pub? pub : sec;
    if (!ctx)
	BUG (0);
    return ctx;
}


/* Get the GPG key with keyid @keyid from the cache. Return it 
   in @r_key on success. */
static gpgme_error_t
get_key_from_cache (const char *keyid, int secret, gpgme_key_t *r_key, 
		    struct keycache_s **c)
{
    gpg_keycache_t cache;
    gpgme_error_t err;
    int mode = secret? KEYCACHE_PRV : KEYCACHE_PUB;

    if (!keyid)
	return gpg_error (GPG_ERR_INV_VALUE);
    if (r_key)
	*r_key = NULL;
    cache = keycache_get_ctx (mode);
    if (!c)
	err = gpg_keycache_find_key (cache, keyid, 0, r_key);
    else
	err = gpg_keycache_find_key2 (cache, keyid, 0, r_key, c);
    return err;
}


/* Release the internal key structure.
   If allocated is 0, assume fixed cache item. */
void
winpt_release_pubkey (winpt_key_s *k)
{
    /*log_box ("debug", 0, "alloc %d", k->allocated);*/
    if (!k->allocated)
	return;
    gpg_keycache_item_release (k->ext);
    k->ext = NULL;
    k->allocated = 0;
}


/* Search the public key with @keyid as the keyid in the cache and
   return the item in @k. */
gpgme_error_t
winpt_get_pubkey (const char *keyid, winpt_key_s *k)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    
    if (pub)
	err = get_key_from_cache (keyid, 0, &k->ctx, &k->ext);
    else
	err = gpg_keycache_fetch_key (keyid, 0, &k->ctx, &k->ext);
    if (err)
	return err;
    k->is_v3 = k->ctx->subkeys->pubkey_algo == GPGME_PK_RSA && 
		strlen (k->ctx->subkeys->fpr) == 32;
    k->is_protected = k->ext->gloflags.is_protected;
    k->keyid = k->ctx->subkeys->keyid+8;
    k->uid = k->ext->uids->uid;
    k->allocated = pub? 0 : 1;
    return 0;
}


gpgme_error_t
winpt_get_seckey (const char *keyid, winpt_key_s *k)
{
    gpgme_error_t err;

    if (sec)
	err = get_key_from_cache (keyid, 1, &k->ctx, &k->ext);
    else
	err = gpg_keycache_fetch_key (keyid, 1, &k->ctx, &k->ext);
    if (err)
	return err;
    k->is_v3 = k->ctx->subkeys->pubkey_algo == GPGME_PK_RSA && 
		strlen (k->ctx->subkeys->fpr) == 32;
    k->is_protected = k->ext->gloflags.is_protected;
    k->keyid = k->ctx->subkeys->keyid+8;
    k->uid = k->ext->uids->uid;
    k->allocated = sec? 0 : 1;
    return 0;
}


gpgme_error_t
get_pubkey (const char *keyid, gpgme_key_t *ret_key)
{
    return get_key_from_cache (keyid, 0, ret_key, NULL);
}


gpgme_error_t
get_seckey (const char *keyid, gpgme_key_t *ret_skey)
{
    return get_key_from_cache (keyid, 1, ret_skey, NULL);
}



/* Map the signature summary in @sum to signature status table index. 
   Return value: index to table. */
static int
sigsum_to_index (gpgme_sigsum_t sum)
{
    if (sum & GPGME_SIGSUM_RED)
	return 2;
    else if (sum & GPGME_SIGSUM_SIG_EXPIRED)
	return 8;
    else if (sum & GPGME_SIGSUM_KEY_REVOKED)
	return 7;
    else if (sum & GPGME_SIGSUM_KEY_EXPIRED)
	return 6;
    else if (sum & GPGME_SIGSUM_GREEN)
	return 1;
    else if (sum & GPGME_SIGSUM_KEY_MISSING)
	return 3;
    return 0;
}


/* Return a humand readable description for the signature status @sum. 
   Warning: this function does not consider the validity of the key. */
const char*
get_gpg_sigstat (gpgme_sigsum_t sum)
{
    const char *gpg_sigstat[] = {
	_("Error during verification process"),
	_("The signature is good"),
	_("The signature is BAD!"),
	_("The signature could not be checked due to a missing key"),
	_("No valid OpenPGP signature"),
	_("Signature Error"),
	_("Good signature (Expired Key)"),
	_("Good signature (Revoked Key)"),
	_("Good signature (Expired)"),
	NULL
    };
    const unsigned int mask = 9;

    return gpg_sigstat[sigsum_to_index (sum) % mask];
}


/* Return true if at least one secret key is available. */
bool
secret_key_available (void)
{
    if (!sec || gpg_keycache_get_size (sec) == 0)
	return false;
    return true;
}
