/* wptKeyManagerDlg.cpp - WinPT Key Manager
 *	Copyright (C) 2000-2006 Timo Schulz
 *	Copyright (C) 2004 Andreas Jobs
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptDlgs.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptKeylist.h"
#include "wptNLS.h"
#include "wptErrors.h"
#include "wptHotkey.h"
#include "wptKeyManager.h"
#include "wptKeyserver.h"
#include "wptKeyEdit.h"
#include "wptRegistry.h"
#include "wptUTF8.h"

/* Name and ID of the separator window. */
#define KM_SEPARATOR_ID			10000
#define WINDOWCLASS_SEPARATOR_CHILD	"WINPT_SEP_CHILD"

/* Virtual key codes. */
#ifndef VK_F
#define VK_F 70
#endif
#ifndef VK_A
#define VK_A 65
#endif
#ifndef VK_C
#define VK_C 67
#endif
#ifndef VK_P
#define VK_P 80
#endif

static subclass_s keylist_proc;

/* Handle to the global image list. */
HIMAGELIST glob_imagelist;

struct km_info_s {
    /* Window positions */
    int pos_x, pos_y;
    int ypos_sep;
    int ypercent_sep;

    /* Different kind of windows. */
    HWND dlg;
    HWND hwnd_sep;
    HWND toolbar;
    HWND statbar;

    listview_ctrl_t lv;
    int		    lv_idx;
    int keylist_sortby;
    int magic;

    unsigned int enable_groups:1;
};
typedef struct km_info_s *km_info_t;

/* Toolbar button structure. */
struct tb_button_s {
    long icon;
    long command;
    char *text;
};

/* The buttons in the toolbar. */
struct tb_button_s tbb[] = {
    {IMI_KEY_DELETE, ID_KEYMISC_DELETE, "Delete"},
    {IMI_KEY_PROPS,  ID_KEYMISC_PROPS,	"Properties"},
    {IMI_KEY_SIGN,   ID_KEYMISC_SIGN,	"Sign"},
    {IMI_KEY_IMPORT, ID_KEYMISC_IMPORT, "Import"},
    {IMI_KEY_EXPORT, ID_KEYMISC_EXPORT, "Export"},
};


static void km_gui_import (struct km_info_s *kmi, int cmd_id, void *param);
void key_get_clip_info (const char *keyid, char *buf, size_t buflen);


/* Subclass the keylist listview control to allow extended commands. */
static BOOL CALLBACK
keylist_subclass_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    struct listview_ctrl_s lv;
    struct km_info_s *kmi;

    switch (msg) {
    case WM_DROPFILES:
	kmi = (km_info_s *)keylist_proc.opaque;
	km_gui_import (kmi, WM_DROPFILES, (void*)wparam);
	break;

    case WM_LBUTTONDBLCLK:
        send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PROPS);
        break;
        
    case WM_KEYUP:
        int virt_key = (int)wparam;
	switch (virt_key) {
	case VK_SPACE:
            send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PROPS);
            break;
        
	case VK_DELETE:
            send_cmd_id (keylist_proc.dlg, ID_KEYMISC_DELETE);
            break;
        
	case VK_INSERT:
	    send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PASTE);
	    break;

	case VK_A:
	    if (GetAsyncKeyState (VK_CONTROL)) {
		lv.ctrl = GetDlgItem (keylist_proc.dlg, IDC_KEYMISC_KEYLIST);
		listview_select_all (&lv);
	    }
	    break;
	    
	case VK_C:
	    if (GetAsyncKeyState (VK_CONTROL)) {
		lv.ctrl = GetDlgItem (keylist_proc.dlg, IDC_KEYMISC_KEYLIST);
		km_clip_export (keylist_proc.dlg, &lv);
	    }
	    break;

	case VK_P:
	    if (GetAsyncKeyState (VK_CONTROL))
		send_cmd_id (keylist_proc.dlg, ID_KEYMISC_PASTE);
	    break;

	case VK_F:
	    if (GetAsyncKeyState (VK_CONTROL)) {
		lv.ctrl = GetDlgItem (keylist_proc.dlg, IDC_KEYMISC_KEYLIST);
		km_find_key (dlg, &lv);
	    }
	    break;
	}
        break;
    }
    
    return CallWindowProc (keylist_proc.old, dlg, msg, wparam, lparam);
}


#define ico2idx(ico) imagelist_getindex ((ico))

/* Create a new toolbar window with a pre-defined list of buttons. */
static HWND
create_toolbar (HWND dlg, struct km_info_s *kmi)
{    
    HWND tbwnd;
    TBSAVEPARAMS tbsp;
    TBBUTTON tb_buttons[] = {
	{ico2idx (IMI_KEY_NEW),	   ID_KEYMISC_KEYWIZARD, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L},
	{ico2idx (IMI_KEY_DELETE), ID_KEYMISC_DELETE, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {ico2idx (IMI_KEY_PROPS),  ID_KEYMISC_PROPS,  TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
	{ico2idx (IMI_KEY_SIGN),   ID_KEYMISC_SIGN,   TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
	{ico2idx (IMI_KEY_SEARCH), ID_KEYMISC_SENDRECV, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {0,                                  0,                 0,               TBSTYLE_SEP,    {0}, 0L, 0},
	{ico2idx (IMI_KEY_FILE_IMPORT), ID_KEYMISC_IMPORT, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {ico2idx (IMI_KEY_FILE_EXPORT), ID_KEYMISC_EXPORT, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
	{ico2idx (IMI_KEY_IMPORT), ID_KEYCTX_PASTE, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
        {ico2idx (IMI_KEY_EXPORT), ID_KEYCTX_COPY, TBSTATE_ENABLED, TBSTYLE_BUTTON, {0}, 0L, 0},
    };

    tbwnd = CreateWindowEx (0, TOOLBARCLASSNAME, NULL,
			    WS_CHILD|TBSTYLE_TOOLTIPS|TBSTYLE_FLAT|CCS_ADJUSTABLE,
			    0, 0, 0, 0, dlg, (HMENU)IDR_WINPT_KMTB, glob_hinst, NULL);
    if (!tbwnd)
	BUG (0);
    SendMessage (tbwnd, TB_BUTTONSTRUCTSIZE, (WPARAM) sizeof(TBBUTTON), 0);
    SendMessage (tbwnd, TB_SETIMAGELIST, 0, (LPARAM)glob_imagelist);
    SendMessage (tbwnd, TB_AUTOSIZE, 0, 0);
    ShowWindow (tbwnd, SW_SHOW);

    /* Restore toolbar */
    tbsp.hkr = HKEY_CURRENT_USER;
    tbsp.pszSubKey = "Software\\WinPT";
    tbsp.pszValueName = "KM_toolbar";
    if (SendMessage (tbwnd, TB_SAVERESTORE, FALSE, (LPARAM)&tbsp) == 0)
        SendMessage (tbwnd, TB_ADDBUTTONS,
		     sizeof (tb_buttons) / sizeof (tb_buttons[0]),
		     (LONG)&tb_buttons[0]);

    return tbwnd;
}


/* Restore the width of the columns from the registry.
   If no bitstring was found, the default size is used. */
int
restore_column_info (struct km_info_s *kmi)
{
    WORD *buf;
    HKEY root;
    DWORD type, size, i;
    LONG ec;

    size = kmi->lv->cols*sizeof (WORD);
    ec = RegOpenKeyEx (HKEY_CURRENT_USER, "Software\\WinPT", 0,
		       KEY_ALL_ACCESS, &root);
    if (ec != ERROR_SUCCESS)
	return -1;

    buf = new WORD[size/2];
    if (!buf)
	BUG (NULL);
    ec = RegQueryValueEx (root, "KMColumnSize", NULL, &type, 
			  (BYTE*)buf, &size);    
    if (ec != ERROR_SUCCESS) {
	RegCloseKey (root);
	free_if_alloc (buf);
	return -1;
    }

    /* check for garbled values. */
    for (i=0; i < size/2; i++) {
	if (buf[i] == 0 || buf[i] > 512) {
	    free_if_alloc (buf);
	    return -1;
	}
    }
    for (i=0; i < size/2; i++)
	listview_set_column_width (kmi->lv, i, buf[i]);
    free_if_alloc (buf);

    size = sizeof (kmi->keylist_sortby);
    ec = RegQueryValueEx (root, "KMSortBy", NULL, &type,
			  (BYTE*)&kmi->keylist_sortby, &size);
    if (ec != ERROR_SUCCESS)
	kmi->keylist_sortby = KEY_SORT_USERID;
    RegCloseKey (root);
    return 0;
}


/* Save the current column width to the registry. */
int
save_column_info (struct km_info_s *kmi)
{    
    HKEY root;
    WORD *buf;
    LONG ec;
    int i;

    buf = new WORD[kmi->lv->cols];
    if (!buf)
	BUG (NULL);
    for (i=0; i < kmi->lv->cols; i++) {
	LVCOLUMN lvc;

	memset (&lvc, 0, sizeof (lvc));
	lvc.mask = LVCF_WIDTH;
	ListView_GetColumn (kmi->lv->ctrl, i, &lvc);
	buf[i] = lvc.cx;
    }

    ec = RegOpenKeyEx (HKEY_CURRENT_USER, "Software\\WinPT", 0, 
		       KEY_ALL_ACCESS, &root);
    if (ec != ERROR_SUCCESS) {
	free_if_alloc (buf);
	return -1;
    }

    ec = RegSetValueEx (root, "KMColumnSize", 0, REG_BINARY,
			(const BYTE*)buf, 2*kmi->lv->cols);
    if (ec == ERROR_SUCCESS) {
	ec = RegSetValueEx (root, "KMSortBy", 0, REG_DWORD_BIG_ENDIAN,
			    (const BYTE*)&kmi->keylist_sortby, 
			    sizeof (kmi->keylist_sortby));
    }
    RegCloseKey (root);
    free_if_alloc (buf);
    return ec == ERROR_SUCCESS? 0 : -1;
}


/* Restore the original size of the dialog. */
static void
restore_window_size (HWND dlg, struct km_info_s *kmi)
{
    char *p;
    int height = 0, width = 0;

    /* Restore the original window size. */
    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Size_Width");
    if (p) {
	width = atoi (p);
	free_if_alloc (p);
    }
    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Size_Height");
    if (p) {
	height = atoi (p);
	free_if_alloc (p);
    }
    if (height > 0 && width > 0)
	SetWindowPos (dlg, HWND_TOP, kmi->pos_x, kmi->pos_y, 
		      height, width, SWP_SHOWWINDOW);
}


/* Center window @dlg. */
static void
do_center_window (HWND dlg, struct km_info_s *kmi)
{
    RECT rect;
    char *p;
    int pos_x = 0;
    int pos_y = 0;
        
    /* Find bottom of keylist */
    GetWindowRect (GetDlgItem(dlg, IDC_KEYMISC_KEYLIST), &rect);
    MapWindowPoints (NULL, dlg, (POINT*)&rect, 2);

    kmi->ypos_sep = rect.bottom;

    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_X");
    if (p && !strcmp (p, " ")) {
	free_if_alloc (p);	
	center_window (dlg, NULL);
	return;
    }
    else if (p)
	pos_x = atoi (p);

    p = get_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_Y");
    if (p && !strcmp (p, " ")) {
	free_if_alloc (p);
	center_window (dlg, NULL);
	return;
    }
    else if (p)
	pos_y = atoi (p);

    if (!pos_y && !pos_x) {
	center_window (dlg, NULL);
	return;
    }
    
    if (pos_x < 0 || pos_y < 0)
	pos_x = pos_y = 0;
    if (pos_x > GetSystemMetrics (SM_CXSCREEN) 
	|| pos_y > GetSystemMetrics (SM_CYSCREEN)) {
	pos_x = pos_y = 0;
    }
    kmi->pos_x = pos_x;
    kmi->pos_y = pos_y;

    GetClientRect (dlg, &rect);
    MoveWindow (dlg, pos_x, pos_y, rect.right, rect.bottom, TRUE);
}


/* Resize the key manager window with the information from @kmi. */
static void
do_resize_window (HWND dlg, struct km_info_s *kmi)
{
    HWND h;
    RECT rclient, rect;
    BOOL bRepaint = FALSE;

    /* Get rect of client area and make life easier */
    GetClientRect (dlg, &rclient);

    /* Move toolbar to the top of the window */
    if (kmi->toolbar) {
        GetWindowRect (kmi->toolbar, &rect);
        ScreenToClient (dlg, (POINT*)&rect);
        ScreenToClient (dlg, (POINT*)&(rect.right));

        rclient.top += rect.bottom - rect.top;
        MoveWindow (kmi->toolbar, 0, 0, rclient.right - rclient.left, 
		    rect.bottom - rect.top, bRepaint);
    }

    /* Move statusbar to the bottom of the window */
    if (kmi->statbar) {
        GetWindowRect (kmi->statbar, &rect);
        ScreenToClient (dlg, (POINT*)&rect);
        ScreenToClient (dlg, (POINT*)&(rect.right));

        rclient.bottom -= rect.bottom - rect.top;
        MoveWindow (kmi->statbar, 0, rclient.bottom,
		    rclient.right - rclient.left,
		    rect.bottom - rect.top, bRepaint);
    }

    /* Obtain separator information and move it to the desired posistion */
    if (kmi->ypercent_sep)
        kmi->ypos_sep = (rclient.bottom - rclient.top) * kmi->ypercent_sep / 100;
    else
        kmi->ypercent_sep = kmi->ypos_sep * 100 / (rclient.bottom - rclient.top);
        
    /* Don't move away */
    if (kmi->ypos_sep+5 > rclient.bottom)
        kmi->ypos_sep = rclient.bottom - 5;
    if (kmi->ypos_sep < rclient.top)
        kmi->ypos_sep = rclient.top;
    MoveWindow (kmi->hwnd_sep, 0, kmi->ypos_sep, 
		(rclient.right - rclient.left), 5, bRepaint);
        
    /* Place the keylist above the separator */
    h = GetDlgItem (dlg, IDC_KEYMISC_KEYLIST);
    MoveWindow (h, rclient.left, rclient.top, rclient.right - rclient.left,
		kmi->ypos_sep - rclient.top, bRepaint);
    rclient.top = kmi->ypos_sep + 5 + 8;

    if (kmi->enable_groups != 0) {
	/* Place the group text and the group box below the separator */
	h = GetDlgItem (dlg, IDC_KEYMISC_GTEXT);
	MoveWindow (h, rclient.left, rclient.top, 100, 14, bRepaint);
	rclient.top += 18;

	h = GetDlgItem (dlg, IDC_KEYMISC_GROUP);
	MoveWindow (h, rclient.left, rclient.top, 
		    rclient.right - rclient.left, 
		    (rclient.bottom < rclient.top) ? 
		    0 : rclient.bottom - rclient.top, bRepaint);
    }
    
    /* Repaint the whole thing */
    InvalidateRect (dlg, NULL, TRUE);
}


/* Return true if the clipboard contains an OpenPGP key. */
static bool
clip_contains_pgpkey (void)
{
    char *ctxt;
    bool val = false;

    ctxt = get_clip_text (NULL);
    if (!ctxt || strlen (ctxt) < 512)
	val = false;
    else if (strstr (ctxt, "BEGIN PGP") && strstr (ctxt, "KEY BLOCK") &&
	     strstr (ctxt, "END PGP"))
	val = true;
    free_if_alloc (ctxt);
    return val;
}


/* Show a mini popup menu to import keys. */
static void
do_create_minpopup (HWND dlg)
{
    HMENU hm;
    POINT p;
    
    if (gpg_read_only || !clip_contains_pgpkey ())
	return;
    hm = CreatePopupMenu ();
    if (!hm)
	BUG (0);
    insert_menu_item (hm, 0, ID_KEYCTX_PASTE, _("Paste Key from Clipboard"));
    GetCursorPos (&p);
    TrackPopupMenu (hm, 0, p.x, p.y, 0, dlg, NULL);
    DestroyMenu (hm);
}


/* Update the default key entry in the status bar for dialog @dlg. */
void
update_default_key_str (HWND dlg)
{
    const char *fmt;
    char *keyid;
    char defkeyinf[128];

    keyid = get_gnupg_default_key ();
    if (!keyid)
	return;
    if ((keyid[0] >= 'A' && keyid[0] <= 'Z') || 
	(keyid[0] >= 'a' && keyid[0] <= 'z') ||
	(keyid[0] == '0' && keyid[1] == 'x'))
        fmt = _("Default Key: %s");
    else
        fmt = _("Default Key: 0x%s");
    _snprintf (defkeyinf, sizeof (defkeyinf) - 1, fmt, keyid);
    SendMessage (dlg, SB_SETTEXT, 0, (LPARAM)defkeyinf);
    gpg_keycache_set_default_key (keycache_get_ctx (0), keyid);
    free_if_alloc (keyid);
}


/* Count all keys and show from @lv results in the status bar @sb. */
void
update_status_bar (HWND sb, listview_ctrl_t lv)
{
    char txt_sec[128];
    char txt_pub[128];
    int nkeys = 0, nsec = 0;

    nkeys = listview_count_items (lv, 0);
    nsec = gpg_keycache_get_size (keycache_get_ctx (0));

    _snprintf (txt_sec, sizeof (txt_sec)-1, _("%d secret keys"), nsec);
    _snprintf (txt_pub, sizeof (txt_pub)-1, _("%d keys"), nkeys);
    SendMessage (sb, SB_SETTEXT, 1, (LPARAM)txt_sec);
    SendMessage (sb, SB_SETTEXT, 2, (LPARAM)txt_pub);
}


long CALLBACK
separator_wnd_proc (HWND hwnd, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static POINT last_mouse_pos;
    PAINTSTRUCT ps;
    RECT rect;
    HPEN hpen;

    if (msg == WM_CREATE)
        SetWindowLong (hwnd, GWL_USERDATA, 
		       (long)(((CREATESTRUCT*)lparam)->lpCreateParams));

    switch (msg) {
    case WM_PAINT:
        GetClientRect (hwnd, &rect);
        BeginPaint (hwnd, &ps);

        /* Background */
        FillRect (ps.hdc, &rect, (HBRUSH)(COLOR_3DFACE+1));

        /* The lines from the light into the dark */
        MoveToEx (ps.hdc, 0,0, NULL);
        hpen = CreatePen (PS_SOLID, 0, GetSysColor (COLOR_3DHILIGHT));
	if (hpen != NULL) {
            SelectObject (ps.hdc, (LPVOID)hpen);
            LineTo (ps.hdc, rect.right, 0);
            DeleteObject (hpen);
        }
	MoveToEx (ps.hdc, 0, 1, NULL);
	hpen = CreatePen (PS_SOLID, 0, GetSysColor (COLOR_3DLIGHT));
	if (hpen != NULL) {
             SelectObject (ps.hdc, (LPVOID)hpen);
             LineTo (ps.hdc, rect.right, rect.bottom);
             DeleteObject (hpen);
         }

        MoveToEx (ps.hdc, 0, rect.bottom-1, NULL);
        hpen = CreatePen (PS_SOLID, 0, GetSysColor(COLOR_3DSHADOW));
	if (hpen != NULL) {
            SelectObject (ps.hdc, (LPVOID)hpen);
            LineTo (ps.hdc, rect.right, rect.bottom-1);
            DeleteObject (hpen);
        }
        MoveToEx (ps.hdc, 0, rect.bottom, NULL);
        hpen = CreatePen(PS_SOLID, 0, GetSysColor(COLOR_3DDKSHADOW));
	if (hpen != NULL) {
            SelectObject (ps.hdc, (LPVOID)hpen);
            LineTo (ps.hdc, rect.right, rect.bottom);
            DeleteObject (hpen);
        }

        EndPaint (hwnd, &ps);
        return 0;

    case WM_LBUTTONDOWN:
        last_mouse_pos.x = LOWORD (lparam);
        last_mouse_pos.y = HIWORD (lparam);
        ClientToScreen (hwnd, &last_mouse_pos);
        SetCapture (hwnd);
        return 0;

    case WM_LBUTTONUP:
        ReleaseCapture ();
        return 0;

    case WM_MOUSEMOVE:
        if (wparam == MK_LBUTTON) {
            struct km_info_s *kmi;
            POINT p;

	    kmi = (struct km_info_s *)GetWindowLong (hwnd, GWL_USERDATA);
            if (kmi == NULL)
                break;

            /* Calculate mouse movement */
            p.x = LOWORD (lparam);
            p.y = HIWORD (lparam);
            ClientToScreen (hwnd, &p);

            GetWindowRect (hwnd, &rect);
            rect.top += (short)(p.y - last_mouse_pos.y);
            rect.bottom += (short)(p.y - last_mouse_pos.y);

            last_mouse_pos.y = p.y;

            /* Apply mouse movement to window. Beware the MoveWindow is relaive 
	       to parent NOT screen */
            MapWindowPoints (NULL, GetParent(hwnd), (POINT*)&rect, 2);
            kmi->ypos_sep = rect.top;
            kmi->ypercent_sep = 0; /* This forces do_resize_window to use abs. position */
	    do_resize_window (GetParent (hwnd), kmi);
	    return 0;
        }
    }

    return DefWindowProc (hwnd, msg, wparam, lparam);
}


/* Register the separator window with @dlg as the parent window. */
static HWND
regist_sep_wnd (HWND dlg, struct km_info_s *kmi)
{
    WNDCLASS wndclass;
    HWND h;

    memset (&wndclass, 0, sizeof (wndclass));
    wndclass.style         = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
    wndclass.lpfnWndProc   = separator_wnd_proc;
    wndclass.cbClsExtra    = 0;
    wndclass.cbWndExtra    = sizeof (long);
    wndclass.hInstance     = glob_hinst;
    wndclass.hIcon         = NULL;
    wndclass.hCursor       = LoadCursor (NULL, IDC_SIZENS);
    wndclass.hbrBackground = (HBRUSH) GetStockObject (LTGRAY_BRUSH);
    wndclass.lpszMenuName  = 0;
    wndclass.lpszClassName = WINDOWCLASS_SEPARATOR_CHILD;
    RegisterClass (&wndclass);

    h = CreateWindowEx (WS_EX_WINDOWEDGE, WINDOWCLASS_SEPARATOR_CHILD, NULL,
			WS_CHILD | WS_CLIPCHILDREN | WS_VISIBLE, 
			0, 400, 2000, 5, dlg, (HMENU) 0, glob_hinst, kmi);
    if (!h)
	BUG (NULL);
    ShowWindow (h, SW_SHOW);
    return h;
}


#define enable_button(hwnd, cid, item_selected) \
    SendMessage ((hwnd), TB_ENABLEBUTTON, (cid), MAKELONG ((item_selected), 0))


/* Interactive modification of the dialog item which depend if an item
   is selected or not and what kind of item. @lv is the list view with
   the items. */
void
update_ui_items (HWND hwnd, listview_ctrl_t lv)
{
    HWND tb_hwnd;
    HMENU hmenu;
    int mult_resids[] = {ID_KEYMISC_PROPS, ID_KEYMISC_SIGN, ID_KEYMISC_EDITKEY,
			 ID_KEYMISC_CHECKSIGS, ID_KEYMISC_REVCERT, 0};
    int key_selected;
    int key_has_sec;
    int key_inv = 0;
    int i, state=0;

    /* Get some states */
    key_selected = SendMessage (GetDlgItem (hwnd, IDC_KEYMISC_KEYLIST), 
					   LVM_GETSELECTEDCOUNT, 0, 0) 
					   ? TRUE : FALSE;
    key_has_sec = FALSE;
    if (key_selected) {
        i = listview_get_curr_pos (lv);
        key_has_sec = km_check_for_seckey (lv, i, NULL) ? TRUE : FALSE;
	key_inv = km_get_key_status (lv, i) & KM_FLAG_REVOKED ||
		  km_get_key_status (lv, i) & KM_FLAG_EXPIRED;
    }

    /* Enable / disable toolbar buttons */
    tb_hwnd = GetDlgItem (hwnd, IDR_WINPT_KMTB);
    enable_button (tb_hwnd, ID_KEYMISC_DELETE, key_selected);
    enable_button (tb_hwnd, ID_KEYMISC_PROPS, key_selected);
    enable_button (tb_hwnd, ID_KEYMISC_SIGN, key_selected && !key_inv);
    enable_button (tb_hwnd, ID_KEYMISC_EXPORT, key_selected);
    enable_button (tb_hwnd, ID_KEYCTX_COPY, key_selected);
    enable_button (tb_hwnd, ID_KEYCTX_PASTE, clip_contains_pgpkey ());

    /* Enable / disable menu items */
    state = key_selected? MF_ENABLED : MF_DISABLED|MF_GRAYED;
    hmenu = GetMenu (hwnd);
    set_menu_state (hmenu, ID_KEYMISC_EXPORT, state);
    set_menu_state (hmenu, ID_KEYMISC_DELETE, state);
    set_menu_state (hmenu, ID_KEYMISC_PROPS, state);
    set_menu_state (hmenu, ID_KEYMISC_EDITKEY, state);
    set_menu_state (hmenu, ID_KEYMISC_CHECKSIGS, state);
    set_menu_state (hmenu, ID_KEYMISC_SIGN, 
		    key_selected && !key_inv ? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_EXPORT_PRIVKEY, 
		    key_selected && key_has_sec? MF_ENABLED : MF_GRAYED);
    set_menu_state (hmenu, ID_KEYMISC_REVCERT, 
		    key_selected && key_has_sec? MF_ENABLED : MF_GRAYED);

    /* Disable some menu items when multiple keys are selected. */
    if (listview_count_items (lv, 1) > 1) {
	for (i=0; mult_resids[i] != 0; i++)
	    set_menu_state (hmenu, mult_resids[i], MF_GRAYED);
    }

    /* Disable all secret-key functions when no secret key is available. */
    if (!secret_key_available ()) {
	enable_button (tb_hwnd, ID_KEYMISC_SIGN, FALSE);
	set_menu_state (hmenu, ID_KEYMISC_SIGN, MF_GRAYED);
    }
}


/* Disable some context menu items when multiple keys are selected. */
static void
popup_multiple (HWND dlg, HMENU hm)
{
    int resids[] = {
	ID_KEYCTX_EDIT,
	ID_KEYCTX_SIGN,
	ID_KEYCTX_REV,
	ID_KEYCTX_ENABLE,
	ID_KEYCTX_DISABLE,
	ID_KEYCTX_ADDKEY,
	ID_KEYCTX_ADDPHOTO,
	ID_KEYCTX_ADDUID,
	ID_KEYCTX_ADDREV,
	ID_KEYCTX_LISTSIGS,
	ID_KEYCTX_MAXTRUST,
	ID_KEYCTX_PROPS,
	ID_KEYCTX_SENDMAIL,
	0};
    int i;
    for (i=0; i < resids[i] != 0; i++)
	set_menu_state (hm, resids[i], MF_GRAYED);
}


/* The items which are shown when GPG is in read-only mode. */
static void
popup_gpg_readonly (HWND dlg, HMENU hm)
{
    int resids[] = {
	ID_KEYCTX_PASTE,
	ID_KEYCTX_EDIT,
	ID_KEYCTX_SIGN,
	ID_KEYCTX_REV,
	ID_KEYCTX_DEL,
	ID_KEYCTX_ENABLE,
	ID_KEYCTX_DISABLE,
	ID_KEYCTX_RECVFROM,
	ID_KEYCTX_ADDKEY,
	ID_KEYCTX_ADDPHOTO,
	ID_KEYCTX_ADDUID,
	ID_KEYCTX_ADDREV,
	0
    };
    int i;

    if (gpg_read_only == 0)
	return;
    for (i=0; resids[i] != 0; i++)
	set_menu_state (hm, resids[i], MF_GRAYED);
}


/* Change the 'Edit' menu based on the current state. */
static void
change_edit_menu (listview_ctrl_t lv, HMENU hm, int id)
{
    enum item { EDIT_MENU = 1 };
    int no_sel;

    if (id != EDIT_MENU)
	return;

    if (!clip_contains_pgpkey ())
	set_menu_state (hm, ID_KEYMISC_PASTE, MF_GRAYED);
    else
	set_menu_state (hm, ID_KEYMISC_PASTE, MF_ENABLED);
    no_sel = listview_get_curr_pos (lv) == -1? 1 : 0;
    set_menu_state (hm, ID_KEYMISC_DELETE2, no_sel? MF_GRAYED: MF_ENABLED);
    set_menu_state (hm, ID_KEYMISC_COPY, no_sel? MF_GRAYED : MF_ENABLED);
}



/* Show limited key menu entries when GPG is in read-only mode. */
static void
change_key_menu (HMENU hm, int id)
{
    int key_resids[] = {
	ID_KEYMISC_SIGN,
	ID_KEYMISC_DELETE,
	ID_KEYMISC_REVCERT,
	ID_KEYMISC_IMPORT_HTTP,
	ID_KEYMISC_IMPORT,
	ID_KEYMISC_REFRESH_KEYS,
	ID_KEYMISC_REBUILD,
	ID_KEYMISC_KEYWIZARD,
	ID_KEYMISC_CARDNEW,
	ID_KEYMISC_NEWKEY,
	0,
    };
    int edit_resids[] = {
	ID_KEYMISC_PASTE,
	ID_KEYMISC_OT,
	ID_KEYMISC_EDITKEY,
	0
    };
    int *resids;
    int i;

    
    if (gpg_read_only == 0)
	return;
    switch (id) {
    case 0: return;
    case 3: resids = key_resids; break;
    case 1: resids = edit_resids;break;
    default:resids = edit_resids; break;
    }

    for (i=0; resids[i] != 0; i++)
	set_menu_state (hm, resids[i], MF_GRAYED);
}


/* Reload a single key in the cache. */
static void
update_key (listview_ctrl_t lv, int pos, const char *keyid, int keypair)
{
    struct winpt_key_s key;

    keycache_update (0, keyid);
    if (keypair)
	keycache_update (1, keyid);

    /* because we replaced the key item, we need to update the
       address of the pointer in the ListView control. */
    memset (&key, 0, sizeof (key));
    winpt_get_pubkey (keyid, &key);
    keylist_upd_key (lv, pos, key.ext, key.ctx);
    keyring_check_last_access ();
}


/* Setup status bar for the main window @dlg. */
static HWND
setup_status_bar (HWND dlg, listview_ctrl_t lv)
{	
    HWND statbar;
    RECT r;
    int partpos[3];
    int i;

    GetClientRect (dlg, &r);
    for (i=1; i <= 3; i++)
	partpos[i-1] = r.right/3*i;
    statbar = CreateStatusWindow (WS_CHILD | WS_VISIBLE | CCS_BOTTOM, 
				  "WinPT_Status_Bar", dlg, KM_SEPARATOR_ID);
    if (!statbar)
	BUG (NULL);
    ShowWindow (statbar, SW_SHOW);
    SendMessage (statbar, SB_SETPARTS, (WPARAM)3, (LPARAM)partpos);

    update_default_key_str (statbar);
    update_status_bar (statbar, lv);
    return statbar;
}


/* Remove or add columns which depends on the state of @checked. */
void
modify_listview_columns (km_info_s *kmi, UINT m_uid, BOOL checked)
{
    UINT resids[] = {
	0,
	ID_KEYMISC_VIEWKEYID,
	ID_KEYMISC_VIEWTYPE,
	0,
	ID_KEYMISC_VIEWCIPHER,
	0,
	0,
	ID_KEYMISC_VIEWCREAT,
	ID_KEYMISC_VIEWDESC,
	-1
    };
    listview_column_s cols[] = {
    {0, 240, (char *)_("User ID")},
    {1, 78, (char *)_("Key ID")},
    {2, 52, (char *)_("Type")},    
    {3, 66, (char *)_("Size")},
    {4, 60, (char *)_("Cipher")},
    {5, 66, (char *)_("Validity")},
    {6, 58, (char *)_("Trust")},
    {7, 72, (char *)_("Creation")},
    {8, 160,(char *)_("Description")},
    {0, 0, NULL}
    };
    UINT pos;

    for (pos=0; resids[pos] != -1; pos++) {
	if (m_uid == resids[pos])
	    break;
    }
    if (!checked)
	listview_del_column (kmi->lv, (int)pos);
    else {	
	listview_add_column (kmi->lv, &cols[pos]);
	keylist_upd_col (kmi->lv, pos);
    }
}


/* Helper to handle the help file. If @check is 1
   the existence of the file is checked. 
   Otherwise the help file will be loaded. */
static bool
start_help (HWND dlg, int check)
{
    DWORD n;
    char path[MAX_PATH+1+32];

    n = GetModuleFileName (NULL, path, sizeof (path)-1-32);
    if (!n)
	return false;
    path[n] = 0;
    while (n-- && path[n] != '\\')
	;
    path[n+1] = 0;
    strcat (path, "winpt.chm");
    if (!check)
	ShellExecute (dlg, "open", path, NULL, NULL, SW_SHOW);
    return file_exist_check (path) == 0? true : false;
}


/* Translate all menu strings. */
static void
translate_menu_strings (HWND dlg)
{
    HMENU menu;

    menu = LoadMenu (glob_hinst, (LPCSTR)IDR_WINPT_KEYMISC);
    set_menu_text_bypos (menu, 0, _("File"));
    set_menu_text_bypos (menu, 1, _("Edit"));
    set_menu_text_bypos (menu, 2, _("View"));
    set_menu_text_bypos (menu, 3, _("Key"));
    set_menu_text_bypos (menu, 4, _("Groups"));

    set_menu_text (menu, ID_KEYMISC_EDITKEY, _("Edit"));
    set_menu_text (menu, ID_KEYMISC_MAIL, _("Send Mail..."));
    set_menu_text (menu, ID_KEYMISC_OT, _("Ownertrust")); /* XXX */
    set_menu_text (menu, ID_KEYMISC_COPY, _("&Copy\tCtrl+C"));
    set_menu_text (menu, ID_KEYMISC_PASTE, _("&Paste\tCtrl+V"));
    set_menu_text (menu, ID_KEYMISC_FIND, _("Search...\tCtrl+F"));
    set_menu_text (menu, ID_KEYMISC_SELALL, _("Select All\tCtrl+A"));
    set_menu_text (menu, ID_KEYMISC_QUIT, _("&Quit"));
    set_menu_text (menu, ID_KEYMISC_UID, _("User ID"));
    set_menu_text (menu, ID_KEYMISC_NEWKEY, _("&Expert"));
    set_menu_text (menu, ID_KEYMISC_KEYWIZARD, _("&Normal"));
    set_menu_text (menu, ID_KEYMISC_EDIT, _("Edit"));
    set_menu_text (menu, ID_KEYMISC_SIGN, _("&Sign"));
    set_menu_text (menu, ID_KEYMISC_DELETE, _("&Delete"));
    set_menu_text (menu, ID_KEYMISC_DELETE2, _("&Delete"));
    set_menu_text (menu, ID_KEYMISC_REVCERT, _("&Revoke Cert"));
    set_menu_text (menu, ID_KEYMISC_CHECKSIGS, _("&List Signatures"));
    set_menu_text (menu, ID_KEYMISC_TRUSTPATH, _("List Trust Path"));
    set_menu_text (menu, ID_KEYMISC_EXPORT, _("&Export..."));
    set_menu_text (menu, ID_KEYMISC_IMPORT, _("&Import..."));
    set_menu_text (menu, ID_KEYMISC_PROPS, _("&Properties"));
    set_menu_text (menu, ID_KEYMISC_GPGOPT, _("GPG &Options"));
    set_menu_text (menu, ID_KEYMISC_GPGPREFS, _("&GPG Preferences"));
    set_menu_text (menu, ID_KEYMISC_WINPTPREFS, _("&WinPT Preferences"));
    set_menu_text (menu, ID_KEYMISC_SENDRECV, _("Keyserver") );
    set_menu_text (menu, ID_KEYMISC_EXPORT_PRIVKEY, _("E&xport Secret Key"));
    set_menu_text (menu, ID_KEYMISC_RECACHE, _("Re&load Key Cache"));
    set_menu_text (menu, ID_KEYMISC_REBUILD, _("R&everify Signatures"));
    set_menu_text (menu, ID_KEYMISC_REFRESH_KEYS, _("Refresh &Keys (Keyserver)"));
    set_menu_text (menu, ID_KEYMISC_INFO, _("Info") );
    set_menu_text (menu, ID_KEYMISC_HELP, _("&Help"));

    set_menu_text (menu, ID_KEYMISC_VIEWKEYID, _("Key ID"));
    set_menu_text (menu, ID_KEYMISC_VIEWCIPHER, _("Cipher"));
    set_menu_text (menu, ID_KEYMISC_VIEWTYPE, _("Type"));
    set_menu_text (menu, ID_KEYMISC_VIEWCREAT, _("Creation"));

    if (!start_help (NULL, 1))
	set_menu_state (menu, ID_KEYMISC_HELP, MF_GRAYED);

    SetMenu (dlg, menu);
}


/* Translate popup menu strings. */
static void
translate_popupmenu_strings (HMENU popup)
{
    set_menu_text (popup, ID_KEYCTX_UID_COPY, _("Copy User ID to Clipboard"));
    set_menu_text (popup, ID_KEYCTX_KEYID_COPY, _("Copy Key ID to Clipboard"));
    set_menu_text (popup, ID_KEYCTX_FPR_COPY, _("Copy Fingerprint to Clipboard"));
    set_menu_text (popup, ID_KEYCTX_KINFO_COPY, _("Copy Key Info to Clipboard"));
    set_menu_text (popup, ID_KEYCTX_EMAIL_COPY, _("Copy E-Mail to Clipboard"));
    set_menu_text (popup, ID_KEYCTX_COPY, _("Copy Key to Clipboard"));
    set_menu_text (popup, ID_KEYCTX_PASTE, _("Paste Key from Clipboard"));
    set_menu_text (popup, ID_KEYCTX_RECVFROM, _("Refresh from Keyserver"));
    set_menu_text (popup, ID_KEYCTX_MAXTRUST, _("Set Implicit &Trust"));
    set_menu_text (popup, ID_KEYCTX_LISTSIGS, _("&List Signatures"));
    set_menu_text (popup, ID_KEYCTX_PROPS, _("&Properties"));
    set_menu_text (popup, ID_KEYCTX_EDIT, _("Key Edit"));
    set_menu_text (popup, ID_KEYCTX_DEL, _("&Delete"));
    set_menu_text (popup, ID_KEYCTX_REV, _("&Revoke Cert"));
    set_menu_text (popup, ID_KEYCTX_SIGN, _("&Sign"));
    set_menu_text (popup, ID_KEYCTX_ENABLE, _("&Enable"));
    set_menu_text (popup, ID_KEYCTX_DISABLE, _("&Disable"));
    set_menu_text (popup, ID_KEYCTX_RECVFROM, _("Re&fresh from Keyserver"));
    set_menu_text (popup, ID_KEYCTX_SETPREFKS, _("Set preferred Keyserver URL"));
    set_menu_text (popup, ID_KEYCTX_SENDMAIL, _("Send Key to Mail Recipient"));
    set_menu_text (popup, ID_KEYCTX_SETDEFKEY, _("Set as Default Key"));

    set_menu_text (popup, ID_KEYCTX_ADDKEY, _("Key..."));
    set_menu_text (popup, ID_KEYCTX_ADDUID, _("User ID..."));
    set_menu_text (popup, ID_KEYCTX_ADDPHOTO, _("Photo ID..."));
    set_menu_text (popup, ID_KEYCTX_ADDREV, _("Revoker..."));

    /* change popup texts */
    set_menu_text_bypos (popup, 0, _("Key Attributes"));
    set_menu_text_bypos (popup, 6, _("Add"));
    set_menu_text_bypos (popup, 19, _("Send to Keyserver"));
}


/* Return true if the cache contain marked keys. */
static bool
updated_keys_avail (void)
{
    gpg_keycache_t kc = keycache_get_ctx (1);
    struct keycache_s *k;

    for (k = kc->item; k; k = k->next) {
	if (k->flags)
	    return true;
    }
    return false;
}


/* Find the index of the key identified by @key. */
static int
find_keypos (listview_ctrl_t lv, gpgme_key_t key)
{
    char keyid[32+1];
    int i;

    for (i=0; i < listview_count_items (lv, 0); i++) {
	listview_get_item_text (lv, i, KM_COL_KEYID, keyid, sizeof (keyid)-1);
	if (!strcmp (key->subkeys->keyid+8, keyid+2))
	    return i;
    }
    return -1;
}


/* Add all recently updated keys in the cache to the list
   and refresh all GUI elements. */
static void
refresh_keylist (struct km_info_s *kmi)
{
    struct keycache_s *ctx;
    gpg_keycache_t kc;
    int status=0, pos;

    kc = keycache_get_ctx (1);
    while (!gpg_keycache_next_updated_key (kc, &ctx, &status)) {
	if (status == KC_FLAG_ADD)
	    keylist_add_key (kmi->lv, KEYLIST_LIST, ctx, ctx->key);
	else {
	    pos = find_keypos (kmi->lv, ctx->key);
	    if (pos != -1)
		keylist_upd_key (kmi->lv, pos, ctx, ctx->key);
	}
    }
    keylist_sort (kmi->lv, kmi->keylist_sortby);
    update_status_bar (kmi->statbar, kmi->lv);
    keyring_check_last_access ();
    SetForegroundWindow (kmi->dlg);
}


/* Reload the key cache. */
static void
keycache_reload (HWND dlg) 
{	
    refresh_cache_s rcs;

    memset (&rcs, 0, sizeof rcs);
    rcs.kr_reload = rcs.kr_update = 1;
    rcs.tr_update = 0;
    DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_KEYCACHE, dlg,					
		    keycache_dlg_proc, (LPARAM)&rcs);
}


/* Check if the secret keyring contains at least one 
   key with ultimate trust.
   Return value: 0 on success. */
int
check_ultimate_trusted_key (gpg_keycache_t _sec)
{
    struct keycache_s *n;
    int nkeys = 0;

    for (n = _sec->item; n; n = n->next) {
	if (n->pubpart && 
	    n->pubpart->key->owner_trust == GPGME_VALIDITY_ULTIMATE)
	    return 0;
	nkeys++;
    }

    /* if we do not have any secret keys, it does not make sense
       to return an error. */
    if (nkeys == 0)
	return 0;
    return -1;
}


/* Reload the entire key cache and update the listview. */
static void
reload_keycache (struct km_info_s *kmi)
{
    gpg_keycache_t c;

    keycache_reload (kmi->dlg);
    c = keycache_get_ctx (1);
    keylist_reload (kmi->lv, c, KEYLIST_LIST, kmi->keylist_sortby);
    update_status_bar (kmi->statbar, kmi->lv);
    SetForegroundWindow (kmi->dlg);
    keyring_check_last_access ();
}


/* Handle all import request. */
static void
km_gui_import (struct km_info_s *kmi, int cmd_id, void *param)
{
    int newkeys = 0, newsks = 0;
    int err = 0;

    switch (cmd_id) {
    case ID_KEYMISC_PASTE:
	err = km_clip_import (kmi->dlg, &newkeys, &newsks);
	break;

    case ID_KEYMISC_IMPORT:
	err = km_file_import (kmi->dlg, NULL, &newkeys, &newsks);
	break;

    case WM_DROPFILES:
	err = km_dropped_file_import (kmi->dlg, (HDROP)param, 
				      &newkeys, &newsks);
	break;

    default:
	err = 1;
	break;
    }

    if (!err && !newsks && (newkeys > 0 && newkeys < KM_PRELOAD_KEYS))
	refresh_keylist (kmi);
    else if (!err) /* avoid to spawn too much processes. */
	reload_keycache (kmi);	
    SetForegroundWindow (kmi->dlg);
}


/* Copy the requested key information (@what) to the clipboard.
   If multiple items were selected, each key info is copied to
   the clipboard. The data is separated by \r\n. */
static void
copy_key_info (int what, struct km_info_s *kmi)
{
    gpgme_data_t out;
    gpgme_key_t key;
    struct keycache_s *kci = NULL;
    size_t n=0;
    char *p, tmp[256];
    int i=0, end = listview_count_items (kmi->lv, 0);
    bool single = false;

    /* Do not iterate over the entire list if just one item is selected. */
    n = listview_count_items (kmi->lv, 1);
    if (n == 1) {
	i = kmi->lv_idx;
	end = i+1;
	single = true;
    }

    gpgme_data_new (&out);
    for (; i < end; i++) {
	if (!listview_get_item_state (kmi->lv, i))
	    continue;
	switch (what) {

	case ID_KEYCTX_EMAIL_COPY:
	    key = km_get_key_ptr (kmi->lv, i, NULL);
	    gpgme_data_write (out, key->uids->email, strlen (key->uids->email));
	    if (!single) gpgme_data_write (out, "\r\n", 2);
	    break;

	case ID_KEYCTX_UID_COPY:
	    key = km_get_key_ptr (kmi->lv, i, &kci);
	    gpgme_data_write (out, kci->uids->uid, strlen (kci->uids->uid));
	    if (!single) gpgme_data_write (out, "\r\n", 2);
	    break;

	case ID_KEYCTX_KEYID_COPY:
	    key = km_get_key_ptr (kmi->lv, i, NULL);
	    gpgme_data_write (out, "0x", 2);
	    gpgme_data_write (out, key->subkeys->keyid+8, 8);
	    if (!single) gpgme_data_write (out, "\r\n", 2);
	    break;

	case ID_KEYCTX_FPR_COPY:
	    key = km_get_key_ptr (kmi->lv, i, NULL);
	    gpgme_data_write (out, key->subkeys->fpr, strlen (key->subkeys->fpr));
	    if (!single) gpgme_data_write (out, "\r\n", 2);
	    break;

	case ID_KEYCTX_KINFO_COPY:
	    key = km_get_key_ptr (kmi->lv, i, NULL);
	    key_get_clip_info (key->subkeys->keyid+8, tmp, sizeof (tmp)-1);
	    gpgme_data_write (out, tmp, strlen (tmp));
	    if (!single) gpgme_data_write (out, "\r\n", 2);
	    break;
	}
    }
    gpgme_data_write (out, "\0", 1);
    p = gpgme_data_release_and_get_mem (out, &n);
    set_clip_text (NULL, p, n);
    gpgme_free (p);
}


/* Dialog box procedure for the Key Manager. */
BOOL CALLBACK
keymanager_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    HWND kl;
    HMENU hm;
    RECT r;
    gpg_keycache_t c;
    gpgme_key_t key;
    km_info_t kmi = NULL;    
    struct genkey_s genkey;
    struct winpt_key_s k = {0};
    struct URL_ctx_s *url;
    char type[32], *name;
    const char *t, *host;
    WORD port = 0;
    int l_idx = 0, i=0, rc;

    if ((msg != WM_INITDIALOG) 
	&& ((kmi = (km_info_t)GetWindowLong (dlg, GWL_USERDATA)) == NULL))
        return FALSE;
    
    /* handle dynamic popup items in the keyserver menu. */
    if (msg == WM_COMMAND && LOWORD (wparam) >= WM_APP && 
			     LOWORD (wparam) <= WM_APP+MAX_KEYSERVERS) {
	l_idx = LOWORD (wparam)-WM_APP;
	if (l_idx < 0)
	    return TRUE;
	host = kserver_get_hostname (l_idx, 0, &port);
	if (host != NULL)
	    km_send_to_keyserver (kmi->lv, dlg, host, port);
	return TRUE;
    }

    switch (msg) {
    case WM_INITDIALOG:
        kmi = new struct km_info_s;
        memset (kmi, 0, sizeof (struct km_info_s));
	kmi->lv_idx = -1;
	kmi->hwnd_sep = regist_sep_wnd (dlg, kmi);
        imagelist_load (dlg);
	translate_menu_strings (dlg);
	SetWindowText (dlg, _("Key Manager"));

        c = keycache_get_ctx (KEYCACHE_PUB);
	kl = GetDlgItem (dlg, IDC_KEYMISC_KEYLIST);
        kmi->keylist_sortby = KEY_SORT_USERID;
        Header_SetImageList(ListView_GetHeader (kl), glob_imagelist);
        kmi->lv = keylist_load (GetDlgItem (dlg, IDC_KEYMISC_KEYLIST), c, 
				NULL, KEYLIST_LIST, kmi->keylist_sortby);
	if (check_ultimate_trusted_key (keycache_get_ctx (0))) {
	    msg_box (dlg, _("No ultimately trusted key found.\n"
			    "Please set at least one secret key to ultimate trust."),
			    _("Key Manager"), MB_WARN);
	}
        /* init subclassing for the listview */
	keylist_proc.opaque = kmi;
        keylist_proc.dlg = dlg;
        keylist_proc.current = (WNDPROC)keylist_subclass_proc;
        keylist_proc.old = (WNDPROC)GetWindowLong(kl, GWL_WNDPROC);
        if (keylist_proc.old) {
            if (!SetWindowLong (kl, GWL_WNDPROC, (LONG)keylist_proc.current)) {
                msg_box (dlg, "Could not set keylist window procedure.",
			 _("Key Manager"), MB_ERR);
		BUG (NULL);
	    }
        }
	kmi->statbar = setup_status_bar (dlg, kmi->lv);
        SetWindowLong (dlg, GWL_USERDATA, (LONG)kmi);
        kmi->toolbar = create_toolbar (dlg, kmi);
	kmi->dlg = dlg;

	do_center_window (dlg, kmi);
        do_resize_window (dlg, kmi);
	restore_window_size (dlg, kmi);
        update_ui_items (dlg, kmi->lv);
	restore_column_info (kmi);
	keylist_sort (kmi->lv, kmi->keylist_sortby);

	SendMessage (dlg, WM_SETICON, (WPARAM)ICON_SMALL, 
		     (LPARAM)LoadIcon (glob_hinst, (LPCTSTR)IDI_WINPT));
	SetDlgItemText (dlg, IDC_KEYMISC_GTEXT, _("Groups"));
	SetDlgItemText (dlg, ID_GROUP_SHOW, _("&Show"));
	SetDlgItemText (dlg, ID_GROUP_NEW, _("&New..."));
	SetForegroundWindow (dlg);
	kmi->magic = SetTimer (dlg, 1, 1000, NULL);
        return TRUE;
   

    case WM_TIMER:
	KillTimer (dlg, kmi->magic);
	SetForegroundWindow (dlg);
	break;
        
    case WM_DESTROY:
	save_column_info (kmi);
	if (kmi->lv) {
            keylist_delete (kmi->lv);
            kmi->lv = NULL;       
	}
	imagelist_destroy ();

	ltoa (kmi->pos_x, type, 10);
	set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_X", type);
	ltoa (kmi->pos_y, type, 10);
	set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Pos_Y", type);

	GetWindowRect (dlg, &r);
	sprintf (type, "%d", (int)(r.bottom-r.top));
	set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Size_Width",
			   type);
	sprintf (type, "%d", (int)(r.right-r.left));
	set_reg_entry (HKEY_CURRENT_USER, "Software\\WinPT", "KM_Size_Height",
			type);

	/* Remove runtime information. 
	   This should be the last action taken here. */
	delete kmi;
	kmi = NULL;
	SetWindowLong (dlg, GWL_USERDATA, 0);
	return FALSE;

    case WM_MOVE:
        GetWindowRect (dlg, &r);
	kmi->pos_x = r.left;
	kmi->pos_y = r.top;	
	break;
        

    case WM_RBUTTONUP:
	do_create_minpopup (dlg);
	break;

    case WM_NOTIFY:		
        NMHDR *notify;
        POINT p;
        HMENU popup;
        
        notify = (NMHDR *)lparam;
	if (!notify)
	    break;
	switch (notify->code) {
        case TBN_QUERYDELETE:
	    SetWindowLong (dlg, DWL_MSGRESULT, TRUE);
            return TRUE;
        
        case TBN_QUERYINSERT:
	    SetWindowLong (dlg, DWL_MSGRESULT, TRUE);
            return TRUE;

        case TBN_GETBUTTONINFO:
            LPTBNOTIFY lpTbNotify;
            lpTbNotify = (LPTBNOTIFY)lparam;
            if (lpTbNotify->iItem < (sizeof(tbb) / sizeof(tb_button_s))) {
                lpTbNotify->tbButton.iBitmap = imagelist_getindex (tbb[lpTbNotify->iItem].icon);
                lpTbNotify->tbButton.idCommand = tbb[lpTbNotify->iItem].command;
                lpTbNotify->tbButton.fsState = TBSTATE_ENABLED;
                lpTbNotify->tbButton.fsStyle = TBSTYLE_BUTTON;
                lpTbNotify->tbButton.iString = 0;
                strncpy (lpTbNotify->pszText, tbb[lpTbNotify->iItem].text, lpTbNotify->cchText);
                SetWindowLong(dlg, DWL_MSGRESULT, TRUE);
            }
            return TRUE;

        case TBN_RESET: /* Restore last saved toolbar status */
            TBSAVEPARAMS tbsp;

            tbsp.hkr = HKEY_CURRENT_USER;
            tbsp.pszSubKey = "Software\\WinPT";
            tbsp.pszValueName = "KM_toolbar";
            SendMessage (notify->hwndFrom, TB_SAVERESTORE, FALSE, (LPARAM)&tbsp); 
            break;

        case TBN_BEGINADJUST: /* User is about to change the toolbar. Save it */
            tbsp.hkr = HKEY_CURRENT_USER;
            tbsp.pszSubKey = "Software\\WinPT";
            tbsp.pszValueName = "KM_toolbar";
            SendMessage (notify->hwndFrom, TB_SAVERESTORE, TRUE, (LPARAM)&tbsp); 
            break;

        case TBN_ENDADJUST: /* User has finished customizing the toolbar. Save it. */
            tbsp.hkr = HKEY_CURRENT_USER;
            tbsp.pszSubKey = "Software\\WinPT";
            tbsp.pszValueName = "KM_toolbar";
            SendMessage (notify->hwndFrom, TB_SAVERESTORE, TRUE, (LPARAM)&tbsp); 
            break;

        case TTN_GETDISPINFO:
            LPTOOLTIPTEXT lpttt;
            lpttt = (LPTOOLTIPTEXT)lparam;

            lpttt->hinst = NULL;
            switch (lpttt->hdr.idFrom) {
	    case ID_KEYMISC_KEYWIZARD:
		lpttt->lpszText = (char*)_("Generate new key pair");
		break;

	    case ID_KEYMISC_SENDRECV:
		lpttt->lpszText = (char*)_("Search for a specific key");
		break;

            case ID_KEYMISC_DELETE:
                lpttt->lpszText = (char *)_("Delete key from keyring"); 
		break;

            case ID_KEYMISC_PROPS:
                lpttt->lpszText = (char *)_("Show key properties"); 
		break;

            case ID_KEYMISC_SIGN:
                lpttt->lpszText = (char *)_("Sign key"); 
		break;

	    case ID_KEYCTX_COPY:
		lpttt->lpszText = (char *)_("Copy key to clipboard");
		break;

	    case ID_KEYCTX_PASTE:
		lpttt->lpszText = (char*)_("Paste key from clipboard");
		break;

            case ID_KEYMISC_IMPORT:
                lpttt->lpszText = (char *)_("Import key to keyring");
		break;

            case ID_KEYMISC_EXPORT:
                lpttt->lpszText = (char *)_("Export key to a file"); 
		break;
            }
            return TRUE;
            
        case LVN_ITEMCHANGED:
            if (((LPNMLISTVIEW)lparam)->uNewState) { /* item selected? */
		kmi->lv_idx = listview_get_curr_pos (kmi->lv);
                update_ui_items (dlg, kmi->lv);
                return TRUE;
            }
            break;

	case NM_RCLICK:
	    if (!wait_for_progress_cb (0))
		break;
            if (notify->idFrom == IDC_KEYMISC_KEYLIST) {
		l_idx =listview_get_curr_pos (kmi->lv);
		if (l_idx == -1)
		    return TRUE; /* Popup only when a item was selected */		
		GetCursorPos (&p);
		hm = LoadMenu (glob_hinst, MAKEINTRESOURCE (IDR_WINPT_KEYMISC_CTX));
		popup = GetSubMenu (hm, 0);
		translate_popupmenu_strings (popup);

		if (km_check_for_seckey (kmi->lv, l_idx, &i))
		    set_menu_state (popup, ID_KEYCTX_SETDEFKEY, MF_ENABLED);
		if (i == 0)
		    set_menu_state (popup, ID_KEYCTX_MAXTRUST, MF_ENABLED);
		if (!km_check_for_seckey (kmi->lv, l_idx, NULL) ||
		    (km_get_key_status (kmi->lv, l_idx) & KM_FLAG_REVOKED)) {
		    set_menu_state (popup, ID_KEYCTX_REV, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDKEY, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDUID, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDREV, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDPHOTO, MF_DISABLED|MF_GRAYED );
		    set_menu_state (popup, ID_KEYCTX_SETPREFKS, MF_DISABLED|MF_GRAYED);
		}
		else if (km_check_for_seckey (kmi->lv, l_idx, NULL) &&
		         km_key_is_v3 (kmi->lv, l_idx)) {
		    /* PGP 2 keys, version 3 have no no support for photo-id's,
		       designated revokers and secondary keys. */
		    set_menu_state (popup, ID_KEYCTX_ADDKEY, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDREV, MF_DISABLED|MF_GRAYED);
		    set_menu_state (popup, ID_KEYCTX_ADDPHOTO, MF_DISABLED|MF_GRAYED);
		}
		if (km_get_key_status( kmi->lv, l_idx ) & KM_FLAG_DISABLED)
		    set_menu_state (popup, ID_KEYCTX_DISABLE, MF_DISABLED|MF_GRAYED);
		else
		    set_menu_state (popup, ID_KEYCTX_ENABLE, MF_DISABLED|MF_GRAYED);
		if (km_get_key_status (kmi->lv, l_idx) & KM_FLAG_REVOKED ||
		    km_get_key_status (kmi->lv, l_idx) & KM_FLAG_EXPIRED)
		    set_menu_state (popup, ID_KEYCTX_SIGN, MF_DISABLED|MF_GRAYED);
		if (!clip_contains_pgpkey ())
		    set_menu_state (popup, ID_KEYCTX_PASTE, MF_DISABLED|MF_GRAYED);
		if (mapi_init ())
		    set_menu_state (popup, ID_KEYCTX_SENDMAIL, MF_DISABLED|MF_GRAYED);
		if (!secret_key_available ())
		    set_menu_state (popup, ID_KEYCTX_SIGN, MF_DISABLED|MF_GRAYED);

		/* Override 'Default Keyserver' with the actual name. */
		host = kserver_get_hostname (0, -1, &port);
		if (!host)
		    host = DEF_HKP_KEYSERVER;
		set_menu_text (popup, ID_KEYCTX_KS_DEFAULT, host);
		{
		    HMENU ks = GetSubMenu (popup, 19);
		    for (i=0; server[i].name != NULL; i++)			
			insert_menu_item (ks, i+2, WM_APP+i, server[i].name);
		}
		popup_gpg_readonly (dlg, popup);
		if (listview_count_items (kmi->lv, 1) > 1)
		    popup_multiple (dlg, popup);
		TrackPopupMenu (popup, TPM_RIGHTALIGN, p.x, p.y, 0, dlg, NULL);
		DestroyMenu (popup);
		DestroyMenu (hm);
		return TRUE;
	    }
	    break;

        case LVN_COLUMNCLICK:
	    if (notify->idFrom == IDC_KEYMISC_KEYLIST) {
                NMLISTVIEW *nft = (LPNMLISTVIEW) lparam;
                int sortby = 0;

		if (!wait_for_progress_cb (0))
		    break;
                switch (nft->iSubItem) {
                case 0:  sortby = KEY_SORT_USERID; break;
                case 1:  sortby = KEY_SORT_KEYID; break;
                case 2:  sortby = KEY_SORT_IS_SECRET; break;
                case 3:  sortby = KEY_SORT_LEN; break;
                case 5:  sortby = KEY_SORT_VALIDITY; break;
		case 6:  sortby = KEY_SORT_OTRUST; break;
                case 7:  sortby = KEY_SORT_CREATED; break;
		case 8:  sortby = KEY_SORT_ALGO; break;
                default: return TRUE; /*sortby userid.*/
                }

                if ((kmi->keylist_sortby & ~KEYLIST_SORT_DESC) == sortby)
                    kmi->keylist_sortby ^= KEYLIST_SORT_DESC;
                else
                    kmi->keylist_sortby = sortby;
                keylist_sort (kmi->lv, kmi->keylist_sortby);
                return TRUE;
            }
            break;
        }
        break;

    case WM_WINDOWPOSCHANGING:
        if (((WINDOWPOS*)lparam)->cx < 400)
            ((WINDOWPOS*)lparam)->cx = 400;
        if (((WINDOWPOS*)lparam)->cy < 200)
            ((WINDOWPOS*)lparam)->cy = 200;
         return TRUE;
        
    case WM_SIZE:
        do_resize_window (dlg, kmi);
        return TRUE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;
        
    case WM_MENUSELECT:
	change_edit_menu (kmi->lv, (HMENU)lparam, LOWORD (wparam));
	change_key_menu ((HMENU)lparam, LOWORD (wparam));
	break;

    case WM_INITMENUPOPUP:
	hm = (HMENU)wparam;
	if ((UINT)LOWORD (lparam) == 3)
	    set_menu_text_bypos (hm, 0, _("New"));
	else if ((UINT)LOWORD (lparam) == 1)
	    set_menu_text_bypos (hm, 8, _("Preferences..."));
	return FALSE;

    case WM_COMMAND:
	/* Allow at least 'Exit' in such a case. */
        if (gnupg_access_keyring (1) && LOWORD (wparam) != ID_KEYMISC_QUIT) {
            msg_box (dlg, _("Could not access public keyring"), 
		     _("Key Manager"), MB_ERR);
            return FALSE;
        }

	/*wait_for_progress_cb ();*/
        switch (LOWORD (wparam)) {
        case ID_KEYMISC_QUIT:
            EndDialog (dlg, TRUE);
            return TRUE;
        
	case ID_KEYMISC_FIND:
	    km_find_key (dlg, kmi->lv);
	    break;

        case ID_KEYMISC_DELETE:
	case ID_KEYMISC_DELETE2:
            if (!km_delete_keys (kmi->lv, dlg))
		update_status_bar (kmi->statbar, kmi->lv);
            return TRUE;
            
        case ID_KEYMISC_SIGN:
	    if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."),  
		         _("Key Manager"), MB_ERR);
                return TRUE;
            }
            if (km_check_key_status (kmi->lv, kmi->lv_idx))
                return TRUE;
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYSIGN, dlg,
                              keysign_dlg_proc, (LPARAM)&k, 
                              _("Key Signing"), IDS_WINPT_KEYSIGN);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 0);
            return TRUE;
            
        case ID_KEYMISC_REVCERT:
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
                
            {
                char state[64];
                listview_get_item_text (kmi->lv, kmi->lv_idx, 5, 
					state, sizeof (state) -1);
                if (strchr (state, 'R' )) {
                    msg_box (dlg, _("Key already revoked!"), 
			     _("Key Manager"), MB_INFO);
                    return TRUE;
                }
            }

	    km_get_key (kmi->lv, kmi->lv_idx, &k);
	    if (!k.key_pair) {
		msg_box (dlg, _("There is no secret key available!"), 
			_("Key Manager"), MB_ERR);
                return TRUE;
            }
            dialog_box_param(glob_hinst, (LPCSTR)IDD_WINPT_KEYREVOKE, dlg,
                             key_revoke_dlg_proc, (LPARAM)&k,
                             _("Key Revocation Cert"), IDS_WINPT_KEYREVOKE);
            return TRUE;
            
        case ID_KEYMISC_TRUSTPATH:
	    if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYTRUST, dlg, 
                              keytrust_dlg_proc, (LPARAM)&k, 
                              _("List Trust Path"), IDS_WINPT_KEYTRUST);
            return TRUE;
            
        case ID_KEYMISC_CHECKSIGS:	    
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYSIG_TREE, dlg, 
                              sigtree_dlg_proc, (LPARAM)&k,
                              _("Key Signature List"), IDS_WINPT_KEYSIG);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 0);
	    if (updated_keys_avail ())
		refresh_keylist (kmi);
            return TRUE;
            
        case ID_KEYMISC_PROPS:	    
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
	    k.callback.ctl = kmi->lv;
	    k.callback.idx = kmi->lv_idx;
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYPROPS, dlg,
                              keyprops_dlg_proc, (LPARAM)&k,
                              _("Key Properties"), IDS_WINPT_KEYPROPS);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, k.key_pair);
            return TRUE;
            
        case ID_KEYMISC_RECACHE:
            if (updated_keys_avail ())
                l_idx = IDYES;
            else {
                l_idx = log_box (_("Key Manager"), MB_YESNO,
                           _("This is only useful when the keyring has been "
                             "modified (sign a key...).\n"
                             "Do you really want to reload the keycache?"));
            }
            if (l_idx == IDYES)
                reload_keycache (kmi);
            return TRUE;
            
        case ID_KEYMISC_REBUILD:
	    name = NULL;
	    gpg_rebuild_cache (&name);
	    if (name != NULL) {
		char *line = strchr (name, '\n');
		show_msg (dlg, 2000, line? name + (line-name)+1 : name);
		safe_free (name);
	    }
	    SetForegroundWindow (dlg);
            return TRUE;
            
        case ID_KEYMISC_NEWKEY:
	    memset (&genkey, 0, sizeof (genkey));
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYGEN, dlg,
                              keygen_dlg_proc, (LPARAM)&genkey, _("Key Generation"),
			      IDS_WINPT_KEYGEN);
	    if (genkey.cancel == 0)
		refresh_keylist (kmi);
            return TRUE;

	case ID_KEYMISC_CARDNEW:
	    if (!scard_support) {
		msg_box (dlg, _("Smart Card support is not available."), 
			 _("Key Manager"), MB_INFO);
		return TRUE;
	    }
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_CARD_KEYGEN, dlg,
			      card_keygen_dlg_proc, 0, _("Card Key Generation"),
			      IDS_WINPT_CARD_KEYGEN);
	    if (updated_keys_avail ())
		send_cmd_id (dlg, ID_KEYMISC_RECACHE);
	    return TRUE;

	case ID_KEYMISC_KEYWIZARD:
	    memset (&genkey, 0, sizeof (genkey));
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYWIZARD, dlg,
			      keygen_wizard_dlg_proc, (LPARAM)&genkey, 
			      _("Key Generation Wizard"),
			      IDS_WINPT_KEYWIZARD);
	    if (genkey.cancel == 0)
		refresh_keylist (kmi);
	    return TRUE;
            
        case ID_KEYMISC_SENDRECV:
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_KEYSERVER, dlg,
                              keyserver_dlg_proc, 0, _("Keyserver Access"),
			      IDS_WINPT_KEYSERVER);
	    refresh_keylist (kmi);
            return TRUE;
            
	case ID_KEYMISC_WINPTPREFS:
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_PREFS,
			      dlg, prefs_dlg_proc, 0,
			      _("WinPT Preferences"), 0);
	    break;

        case ID_KEYMISC_GPGPREFS:
            rc = dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_GPGPREFS, 
				   dlg, gpgprefs_dlg_proc, 0, 
				   _("GnuPG Preferences"), IDS_WINPT_GPGPREFS);
	    if (rc == TRUE) {
		reload_keycache (kmi);
		update_default_key_str (kmi->statbar);
	    }
            return TRUE;
            
        case ID_KEYMISC_GPGOPT:
            dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_GPGOPT, dlg,
                              gpgopt_dlg_proc, 0, _("GnuPG Options"), 
			      IDS_WINPT_GPGOPT);
            return TRUE;
            
        case ID_KEYMISC_IMPORT:
            km_gui_import (kmi, LOWORD (wparam), NULL);
            return TRUE;

	case ID_KEYMISC_IMPORT_HTTP:
	    url = (struct URL_ctx_s*)get_http_file_dlg (dlg);
	    if (url && url->cancel == 0) {
		km_http_import (dlg, url->url);
		refresh_keylist (kmi);
	    }
	    free_if_alloc (url);
	    break;
            
        case ID_KEYMISC_EXPORT:
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."),
			 _("Key Manager"), MB_ERR);
                return TRUE;
            }
	    if (listview_count_items (kmi->lv, 1) > 1)
		name = m_strdup ("Exported_GPG_Keys.asc");
	    else {
		key = km_get_key_ptr (kmi->lv, kmi->lv_idx, NULL);
		name = km_gen_export_filename (key->subkeys->keyid+8, 0);
	    }
            t = get_filesave_dlg (dlg, _("Choose Name for Key File"), NULL, name);
	    free_if_alloc (name);
            if (t == NULL)
                return TRUE;
            km_file_export (dlg, kmi->lv, t);
            return TRUE;
            
        case ID_KEYMISC_EXPORT_PRIVKEY:
            if (kmi->lv_idx == -1) {
                msg_box (dlg, _("Please select a key."), _("Key Manager"), MB_ERR);
                return TRUE;
            }
            if( !km_check_for_seckey( kmi->lv, kmi->lv_idx, NULL ) ) {
                msg_box (dlg, _("There is no corresponding secret key for this key."),
                         _("Key Manager"), MB_ERR);
                return TRUE;
            }
            if (listview_count_items (kmi->lv, 1) > 1) {
                msg_box (dlg, _("You can only export one secret key."), 
			 _("Key Manager"), MB_ERR);
                return TRUE;
            }
            i = msg_box (dlg, 
                          _("This operation will export your *SECRET* key!\n\n"
                            "Never send this key to ANYONE, it should be available\n"
                            "ONLY on your machine and you may use this function\n"
                            "to copy the key to a safe place.\n\n"
                            "Do you really want to export the key?"),
                          _("WARNING"), MB_INFO|MB_YESNO);
            if (i == IDYES) {
		key = km_get_key_ptr (kmi->lv, kmi->lv_idx, NULL);
		name = km_gen_export_filename (key->subkeys->keyid+8, 1);
                t = get_filesave_dlg (dlg, _("Choose Name for Key File"), NULL, name);
                if (t != NULL)		
		    km_privkey_export (dlg, kmi->lv, t);
            }
            return TRUE;

	case ID_KEYMISC_INFO:
	    dialog_box_param (glob_hinst, (LPCSTR)IDD_WINPT_ABOUT, dlg,
			      about_winpt_dlg_proc, 0, _("About WinPT"),
			      IDS_WINPT_ABOUT);
	    break;

	case ID_KEYMISC_DEVWWW:
	    ShellExecute (dlg, "open", "http://wald.intevation.org/projects/winpt",
			  NULL, NULL, SW_SHOW);
	    break;

	case ID_KEYMISC_WEBSITE:
	    ShellExecute (dlg, "open", "http://www.winpt.org", 
			  NULL, NULL, SW_SHOW);
	    break;

	case ID_KEYMISC_HELP:
	    start_help (dlg, 0);
	    break;

	case ID_KEYMISC_OT:
	    dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_OWNERTRUST, 
			      dlg, ownertrust_dlg_proc, 0,
			      _("Ownertrust"), IDS_WINPT_OWNERTRUST);
	    break;

	case ID_KEYMISC_EDITKEY:
	    if (km_get_key (kmi->lv, kmi->lv_idx, &k))
		break;
	    dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_KEYEDIT, dlg,
			      keyedit_main_dlg_proc, (LPARAM)&k,
			      _("Key Edit"), IDS_KEYCTX_EDIT);
	    if (k.update)
		update_key (kmi->lv,  kmi->lv_idx, k.tmp_keyid, 1);
	    break;
            
        case ID_KEYMISC_COPY:
            km_clip_export (dlg, kmi->lv);
            break;
            
	case ID_KEYMISC_SELALL:
	    listview_select_all (kmi->lv);
	    break;

        case ID_KEYMISC_PASTE:
	    km_gui_import (kmi, LOWORD (wparam), NULL);
            break;
            
	case ID_KEYCTX_SETPREFKS:
	    if (km_get_key (kmi->lv, kmi->lv_idx, &k))
		break;
	    keyedit_set_pref_keyserver (&k, dlg);
	    break;

        case ID_KEYMISC_REFRESH_KEYS:
	    if (listview_count_items (kmi->lv, 1) == 0) {
		msg_box (dlg, _("No key was selected, select all by default."), 
			 _("Key Manager"), MB_INFO);
		listview_select_all (kmi->lv);
	    }
            if (!km_refresh_from_keyserver (kmi->lv, dlg))
		msg_box (dlg, _("Keyserver refresh finished."), 
			 _("Key Manager"), MB_OK);
            break;
                        
	/** Context menu **/
	case ID_KEYCTX_EDIT:
	    send_cmd_id( dlg, ID_KEYMISC_EDITKEY );
	    break;

        case ID_KEYCTX_PROPS:
	    send_cmd_id( dlg, ID_KEYMISC_PROPS );
	    break;

        case ID_KEYCTX_UIDS: 
	    send_cmd_id( dlg, ID_KEYMISC_UID ); 
	    break;

        case ID_KEYCTX_SIGN: 
	    send_cmd_id( dlg, ID_KEYMISC_SIGN ); 
	    break;

        case ID_KEYCTX_DEL: 
	    send_cmd_id (dlg, ID_KEYMISC_DELETE); 
	    break;

        case ID_KEYCTX_REV: 
	    send_cmd_id( dlg, ID_KEYMISC_REVCERT ); 
	    break;

	case ID_KEYCTX_SENDMAIL:
	    km_send_to_mail_recipient (kmi->lv, dlg);
	    break;

        case ID_KEYCTX_KS_DEFAULT:
	    host = kserver_get_hostname (0, -1, &port);
	    km_send_to_keyserver (kmi->lv, dlg, host, port);
	    break;

	case ID_KEYCTX_ADDKEY:
            km_get_key (kmi->lv, kmi->lv_idx, &k);
	    keyedit_add_subkey (&k, dlg, NULL);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 1);
	    break;

	case ID_KEYCTX_ADDUID:
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
	    keyedit_add_userid (&k, dlg, NULL);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 1);
	    break;

	case ID_KEYCTX_ADDREV:
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
	    keyedit_add_revoker (&k, dlg);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 1);
	    break;

	case ID_KEYCTX_ADDPHOTO:
	    km_get_key (kmi->lv, kmi->lv_idx, &k);
	    keyedit_add_photo (&k, dlg);
	    if (k.update)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 1);
	    break;

        case ID_KEYCTX_RECVFROM:
	    km_refresh_from_keyserver (kmi->lv, dlg);
	    if (updated_keys_avail ())
		refresh_keylist (kmi);
	    break;

	case ID_KEYCTX_UID_COPY:
	case ID_KEYCTX_KEYID_COPY:
	case ID_KEYCTX_FPR_COPY:
	case ID_KEYCTX_KINFO_COPY:
	case ID_KEYCTX_EMAIL_COPY:
	    copy_key_info (LOWORD (wparam), kmi);
	    break;

        case ID_KEYCTX_COPY:
	    send_cmd_id (dlg, ID_KEYMISC_COPY);
            break;

        case ID_KEYCTX_PASTE:
            send_cmd_id (dlg, ID_KEYMISC_PASTE);
            break;

	case ID_KEYCTX_DISABLE:
	case ID_KEYCTX_ENABLE:
	    i = LOWORD (wparam) == ID_KEYCTX_ENABLE? 1 : 0;
	    if (km_get_key (kmi->lv, kmi->lv_idx, &k))
		break;
	    rc = km_enable_disable_key (kmi->lv, dlg, kmi->lv_idx, i);
	    if (!rc)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 0);
	    break;

        case ID_KEYCTX_LISTSIGS: 
	    send_cmd_id (dlg, ID_KEYMISC_CHECKSIGS);
	    break;

	case ID_KEYCTX_MAXTRUST:
	    if (km_get_key (kmi->lv, kmi->lv_idx, &k))
		break;
	    rc = km_set_implicit_trust (dlg, kmi->lv, kmi->lv_idx);
	    if (!rc)
		update_key (kmi->lv, kmi->lv_idx, k.tmp_keyid, 0);
	    break;

        case ID_KEYCTX_SETDEFKEY:
	    if (!km_check_key_status (kmi->lv, kmi->lv_idx)) {
		key = km_get_key_ptr (kmi->lv, kmi->lv_idx, NULL);
		rc = set_gnupg_default_key (key->subkeys->keyid+8);
		if (rc)
		    msg_box (dlg, winpt_strerror (rc), _("Key Manager"), MB_ERR);
		update_default_key_str (kmi->statbar);
	    }
            break;

	case ID_KEYMISC_VIEWKEYID:
	case ID_KEYMISC_VIEWCIPHER:
	case ID_KEYMISC_VIEWTYPE:
	case ID_KEYMISC_VIEWCREAT:
	case ID_KEYMISC_VIEWDESC:
	    hm = GetMenu (dlg);
	    i = get_menu_state (hm, LOWORD (wparam));
	    set_menu_state (hm, LOWORD (wparam), 
			    i & MFS_CHECKED? MFS_UNCHECKED : MFS_CHECKED);
	    modify_listview_columns (kmi, LOWORD (wparam), !(i & MFS_CHECKED));
	    break;

	case ID_GROUP_NEW:
        case ID_GROUP_PASTE:
        case ID_GROUP_DELETE:
	case ID_GROUP_SHOW:
	    /* XXX: Implement group manager. */
            return TRUE;
        }
	
        break;
    }
    
    return FALSE;
}
